"""
Copyright 2022-2025, CCL Forensics

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""


import abc
import dataclasses
import io
import os
import re
import sys
import types
import typing
import pathlib
import datetime
import struct
import enum
import zlib

__version__ = "0.22"
__description__ = "Library for reading Chrome/Chromium Cache (both blockfile and simple format)"
__contact__ = "Alex Caithness"


_CHROME_EPOCH = datetime.datetime(1601, 1, 1)
EIGHT_BYTE_PICKLE_ALIGNMENT = True  # switch this if you get errors about the EOF magic when doing a Simple Cache
SIMPLE_EOF_SIZE = 24 if EIGHT_BYTE_PICKLE_ALIGNMENT else 20


def decode_chrome_time(us: int) -> datetime.datetime:
    return _CHROME_EPOCH + datetime.timedelta(microseconds=us)


class BinaryReader:
    """
    Utility class which wraps a BinaryIO and provides reading for a bunch of data types we need to do the cache stuff
    """
    def __init__(self, stream: typing.BinaryIO):
        self._stream = stream
        self._closed = False

    @classmethod
    def from_bytes(cls, buffer: bytes):
        return cls(io.BytesIO(buffer))

    def close(self):
        self._stream.close()
        self._closed = True

    def __enter__(self) -> "BinaryReader":
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def tell(self) -> int:
        return self._stream.tell()

    def seek(self, offset: int, whence: int) -> int:
        return self._stream.seek(offset, whence)

    def read_raw(self, count: int) -> bytes:
        start_offset = self._stream.tell()
        result = self._stream.read(count)
        if len(result) != count:
            raise ValueError(
                f"Could not read all of the data starting at {start_offset}. Expected: {count}; got {len(result)}")
        return result

    def read_utf8(self, count: int) -> str:
        return self.read_raw(count).decode("utf-8")

    def read_int16(self) -> int:
        raw = self.read_raw(2)
        return struct.unpack("<h", raw)[0]

    def read_int32(self) -> int:
        raw = self.read_raw(4)
        return struct.unpack("<i", raw)[0]

    def read_int64(self) -> int:
        raw = self.read_raw(8)
        return struct.unpack("<q", raw)[0]

    def read_uint16(self) -> int:
        raw = self.read_raw(2)
        return struct.unpack("<H", raw)[0]

    def read_uint32(self) -> int:
        raw = self.read_raw(4)
        return struct.unpack("<I", raw)[0]

    def read_uint64(self) -> int:
        raw = self.read_raw(8)
        return struct.unpack("<Q", raw)[0]

    def read_addr(self) -> "Addr":
        return Addr.from_int(self.read_uint32())

    def read_datetime(self) -> datetime.datetime:
        return decode_chrome_time(self.read_uint64())

    @property
    def is_closed(self) -> bool:
        return self._closed

    @property
    def is_eof(self) -> bool:
        test = self._stream.read(1)
        if len(test) == 0:
            return True
        self._stream.seek(-1, os.SEEK_CUR)
        return False


class FileType(enum.IntEnum):
    # net/disk_cache/blockfile/disk_format.h
    EXTERNAL = 0
    RANKINGS = 1
    BLOCK_256 = 2
    BLOCK_1K = 3
    BLOCK_4K = 4
    BLOCK_FILES = 5
    BLOCK_ENTRIES = 6
    BLOCK_EVICTED = 7


_BLOCKSIZE_FOR_FILETYPE = {
    FileType.RANKINGS: 36,
    FileType.BLOCK_256: 256,
    FileType.BLOCK_1K: 1024,
    FileType.BLOCK_4K: 4096,
    FileType.BLOCK_FILES: 8,
    FileType.BLOCK_ENTRIES: 104,
    FileType.BLOCK_EVICTED: 48,
    FileType.EXTERNAL: 0
}


_BLOCK_FILE_FILETYPE = {FileType.BLOCK_256, FileType.BLOCK_1K, FileType.BLOCK_4K}


class CacheKey:
    """
    Class representing a parsed Chromium Cache Key.
    """
    # net/http/http_cache.cc GenerateCacheKey
    CRED_UPLOAD_KEY_PREFIX_PATTERN = re.compile(r"^\d+/\d+/")  # 'current' (since Sept '21)
    UPLOAD_ONLY_KEY_PREFIX_PATTERN = re.compile(r"^\d+/")  # prior to Sept '21
    # if neither of the above we assume we only have a URL

    def __init__(self, raw_key: str):
        self._raw_key = raw_key

        # We have to account for a few different versions of keys, we can do this based on a prefix
        if CacheKey.UPLOAD_ONLY_KEY_PREFIX_PATTERN.match(self._raw_key):
            if CacheKey.CRED_UPLOAD_KEY_PREFIX_PATTERN.match(self._raw_key):
                split_key = self._raw_key.split("/", 2)
                self._credential_key = split_key[0]
                self._upload_data_identifier = int(split_key[1])
            else:
                split_key = self._raw_key.split("/", 1)
                self._credential_key = ""
                self._upload_data_identifier = int(split_key[0])

            if split_key[-1].startswith("_dk_"):
                # consume two kDoubleKeySeparator (a space), the url is after that
                (self._isolation_key_top_frame_site,
                 self._isolation_key_variable_part,
                 self._url) = split_key[-1][4:].split(" ", 3)
                if self._isolation_key_top_frame_site.startswith("s_"):
                    self._isolation_key_top_frame_site = self._isolation_key_top_frame_site[2:]
            else:
                self._url = split_key[-1]
                self._isolation_key_top_frame_site = None
                self._isolation_key_variable_part = None
        else:
            # if the prefixes don't hit, this should just be a URL
            self._url = self._raw_key
            self._isolation_key_top_frame_site = None
            self._isolation_key_variable_part = None

    @property
    def raw_key(self) -> str:
        return self._raw_key

    @property
    def url(self) -> str:
        return self._url

    @property
    def credential_key(self) -> str:
        return self._credential_key

    @property
    def upload_data_identifier(self) -> int:
        return self._upload_data_identifier

    @property
    def isolation_key_top_frame_site(self) -> str:
        return self._isolation_key_top_frame_site

    @property
    def isolation_key_variable_part(self):
        return self._isolation_key_variable_part

    def __str__(self):
        return self._raw_key

    def __repr__(self):
        return (f"<CacheKey url: {self._url}; credential_key: {self._credential_key}; "
                f"upload_data_identifier: {self._upload_data_identifier}; "
                f"isolation_key_top_frame_site: {self._isolation_key_top_frame_site}; "
                f"isolation_key_variable_part: {self._isolation_key_variable_part}>")


class Addr:
    # net/disk_cache/blockfile/addr.h
    def __init__(
            self, is_initialized: bool, file_type: FileType, file_number: typing.Optional[int],
            contiguous_blocks: typing.Optional[int], file_selector: typing.Optional[int], block_number: int,
            reserved_bits: typing.Optional[int]):
        self._is_initialized = is_initialized
        self._file_type = file_type
        self._file_number = file_number
        self._contiguous_blocks = contiguous_blocks
        self._file_selector = file_selector
        self._block_number = block_number
        self._reserved_bits = reserved_bits

    def __repr__(self):
        return (f"<Addr: is_initialized: {self._is_initialized}; file_type: {self._file_type.name}; "
                f"file_number: {self._file_number}; contiguous_blocks: {self._contiguous_blocks}; "
                f"file_selector: {self._file_selector}; block_number: {self._block_number}>")

    @classmethod
    def from_int(cls, i: int):
        is_initialized = (i & 0x80000000) > 0
        file_type = FileType((i & 0x70000000) >> 28)

        if file_type == FileType.EXTERNAL:
            file_number = i & 0x0fffffff
            contiguous_blocks = None
            file_selector = None
            block_number = None
            reserved_bits = None
        else:
            file_number = None
            contiguous_blocks = 1 + ((i & 0x03000000) >> 24)
            file_selector = (i & 0x00ff0000) >> 16
            block_number = i & 0x0000ffff
            reserved_bits = i & 0x0c000000

        return Addr(
            is_initialized,
            file_type,
            file_number,
            contiguous_blocks,
            file_selector,
            block_number,
            reserved_bits)

    def sanity_check(self) -> bool:
        # implementation from addr.cc - will hopefully identify invalid data and skip it rather than raising exceptions
        #  we omit the initialized check from that version, as that's to identify a totally empty entry (which is sane
        #  but of no use to us in any context we use it).
        if self._file_type > FileType.BLOCK_4K:
            return False
        if self._file_type != FileType.EXTERNAL and self._reserved_bits != 0:
            return False

        return True

    def sanity_check_for_entry(self) -> bool:
        return self.sanity_check() and self._file_type == FileType.BLOCK_256

    @property
    def is_initialized(self) -> bool:
        return self._is_initialized

    @property
    def file_type(self) -> FileType:
        return self._file_type

    @property
    def contiguous_blocks(self) -> int:
        return self._contiguous_blocks

    @property
    def file_selector(self) -> int:
        return self._file_selector

    @property
    def block_number(self) -> int:
        return self._block_number

    @property
    def external_file_number(self) -> int:
        return self._file_number


@dataclasses.dataclass(frozen=True)
class LruData:
    # net/disk_cache/blockfile/disk_format.h
    filled: int
    sizes: typing.Collection[int]
    heads: typing.Collection[Addr]
    tails: typing.Collection[Addr]
    transactions: Addr
    operation: int
    operation_list: int

    @classmethod
    def from_bytes(cls, buffer: bytes):
        with BinaryReader.from_bytes(buffer) as reader:
            return cls.from_reader(reader)

    @classmethod
    def from_reader(cls, reader: BinaryReader):
        _ = [reader.read_int32() for x in range(2)]
        filled = reader.read_int32()
        sizes = tuple(reader.read_int32() for _ in range(5))
        heads = tuple(reader.read_addr() for _ in range(5))
        tails = tuple(reader.read_addr() for _ in range(5))
        transaction = reader.read_addr()
        operation = reader.read_int32()
        operation_list = reader.read_int32()
        _ = [reader.read_int32() for x in range(7)]

        return cls(filled, sizes, heads, tails, transaction, operation, operation_list)


@dataclasses.dataclass(frozen=True)
class BlockFileIndexHeader:
    # net/disk_cache/blockfile/disk_format.h
    version: int
    num_entries: int
    num_bytes_v2: int
    last_file: int
    this_id: int
    stats_addr: Addr
    table_length: int
    crash: int
    experiment: int
    create_time: datetime.datetime
    num_bytes_v3: int
    lru: LruData

    _MAGIC: typing.ClassVar[int] = 0xC103CAC3

    @classmethod
    def from_bytes(cls, buffer: bytes):
        with BinaryReader.from_bytes(buffer) as reader:
            return cls.from_reader(reader)

    @classmethod
    def from_reader(cls, reader: BinaryReader):
        magic = reader.read_uint32()
        if magic != BlockFileIndexHeader._MAGIC:
            raise ValueError("invalid magic")
        version = reader.read_uint32()
        num_entries = reader.read_int32()
        old_v2_num_bytes = reader.read_uint32()
        last_file = reader.read_int32()
        this_id = reader.read_int32()
        stats_addr = reader.read_addr()
        table_length = reader.read_int32() or 0x10000
        crash = reader.read_int32()
        experiment = reader.read_int32()
        create_time = reader.read_datetime()
        num_bytes = reader.read_int64()
        _ = [reader.read_int32() for x in range(50)]
        lru = LruData.from_reader(reader)

        return cls(
            version, num_entries, old_v2_num_bytes, last_file, this_id, stats_addr,
            table_length, crash, experiment, create_time, num_bytes, lru)


class BlockFileIndexFile:
    # net/disk_cache/blockfile/disk_format.h
    def __init__(self, file_path: typing.Union[os.PathLike, str]):
        self._input_path = pathlib.Path(file_path)
        with BinaryReader(self._input_path.open("rb")) as reader:
            self._header = BlockFileIndexHeader.from_reader(reader)
            self._entries = tuple(reader.read_addr() for _ in range(self._header.table_length))
            self._entries_initialized = tuple(x for x in self._entries if x.is_initialized)

    @property
    def input_path(self):
        return self._input_path

    @property
    def header(self) -> BlockFileIndexHeader:
        return self._header

    @property
    def index(self) -> typing.Collection[Addr]:
        return self._entries

    @property
    def index_initialized_only(self):
        return self._entries_initialized


class EntryState(enum.IntEnum):
    NORMAL = 0
    EVICTED = 1
    DOOMED = 2


class EntryFlags(enum.IntFlag):
    PARENT_ENTRY = 1 << 0
    CHILD_ENTRY = 1 << 1


@dataclasses.dataclass(frozen=True)
class EntryStore:
    # net/disk_cache/blockfile/disk_format.h
    entry_hash: int
    next_entry: Addr
    rankings_node: Addr
    reuse_count: int
    refetch_count: int
    state: EntryState
    creation_time: datetime.datetime
    key_length: int
    long_key_addr: Addr
    data_sizes: tuple[int, int, int, int]
    data_addrs: tuple[Addr, Addr, Addr, Addr]
    flags: EntryFlags
    self_hash: int
    key: typing.Optional[str]

    @property
    def key_is_external(self) -> bool:
        return self.long_key_addr.is_initialized

    @classmethod
    def from_bytes(cls, buffer: bytes):
        with BinaryReader.from_bytes(buffer) as reader:
            return cls.from_reader(reader)

    @classmethod
    def from_reader(cls, reader: BinaryReader):
        start = reader.tell()

        entry_hash = reader.read_uint32()
        next_entry = reader.read_addr()
        rankings_node = reader.read_addr()
        reuse_count = reader.read_int32()
        refetch_count = reader.read_int32()
        state = EntryState(reader.read_int32())
        creation_time = reader.read_datetime()
        key_length = reader.read_int32()
        long_key_addr = reader.read_addr()
        data_sizes = (reader.read_int32(), reader.read_int32(), reader.read_int32(), reader.read_int32())
        data_addrs = (reader.read_addr(), reader.read_addr(), reader.read_addr(), reader.read_addr())
        flags = EntryFlags(reader.read_uint32())
        _ = [reader.read_int32() for x in range(4)]
        self_hash = reader.read_uint32()

        meta_length = reader.tell() - start

        key = None
        key_is_external = long_key_addr.is_initialized
        if not key_is_external:
            key = reader.read_utf8(key_length)

        return cls(
            entry_hash, next_entry, rankings_node, reuse_count, refetch_count, state, creation_time, key_length,
            long_key_addr, data_sizes, data_addrs, flags, self_hash, key)


@dataclasses.dataclass(frozen=True)
class BlockFileHeader:
    # net/disk_cache/blockfile/disk_format_base.h
    version: int
    this_file: int
    next_file: int
    entry_size: int
    num_entries: int
    max_entries: int
    empty_type_counts: tuple[int, int, int, int]
    hints: tuple[int, int, int, int]
    updating: int
    user: tuple[int, int, int, int, int]
    allocation_map: bytes

    _MAGIC: typing.ClassVar[int] = 0xC104CAC3
    _BLOCK_HEADER_SIZE: typing.ClassVar[int] = 8192
    _MAX_BLOCKS: typing.ClassVar[int] = (_BLOCK_HEADER_SIZE - 80) * 8

    def __post_init__(self):
        if len(self.allocation_map) != self._MAX_BLOCKS // 8:
            raise ValueError("invalid allocation map length")

    @classmethod
    def from_bytes(cls, buffer: bytes):
        with BinaryReader.from_bytes(buffer) as reader:
            return cls.from_reader(reader)

    @classmethod
    def from_reader(cls, reader: BinaryReader):
        magic = reader.read_uint32()
        if magic != cls._MAGIC:
            raise ValueError("invalid magic")
        version = reader.read_uint32()
        this_file = reader.read_int16()
        next_file = reader.read_int16()
        entry_size = reader.read_int32()
        num_entries = reader.read_int32()
        max_entries = reader.read_int32()
        empty = (reader.read_int32(), reader.read_int32(), reader.read_int32(), reader.read_int32())
        hints = (reader.read_int32(), reader.read_int32(), reader.read_int32(), reader.read_int32())
        updating = reader.read_int32()
        user = (reader.read_int32(), reader.read_int32(), reader.read_int32(), reader.read_int32(), reader.read_int32())

        allocation_map = reader.read_raw(cls._MAX_BLOCKS // 8)

        return cls(
            version, this_file, next_file, entry_size, num_entries, max_entries,
            empty, hints, updating, user, allocation_map)


class CachedMetadataFlags(enum.IntFlag):
    # net/http/http_response_info.cc

    RESPONSE_INFO_VERSION = 3
    RESPONSE_INFO_VERSION_MASK = 0xFF

    # This bit is set if the response info has a cert at the end.
    RESPONSE_INFO_HAS_CERT = 1 << 8
    RESPONSE_INFO_HAS_SECURITY_BITS = 1 << 9
    RESPONSE_INFO_HAS_CERT_STATUS = 1 << 10
    RESPONSE_INFO_HAS_VARY_DATA = 1 << 11
    RESPONSE_INFO_TRUNCATED = 1 << 12
    RESPONSE_INFO_WAS_SPDY = 1 << 13
    RESPONSE_INFO_WAS_ALPN = 1 << 14
    RESPONSE_INFO_WAS_PROXY = 1 << 15
    RESPONSE_INFO_HAS_SSL_CONNECTION_STATUS = 1 << 16
    RESPONSE_INFO_HAS_ALPN_NEGOTIATED_PROTOCOL = 1 << 17
    RESPONSE_INFO_HAS_CONNECTION_INFO = 1 << 18
    RESPONSE_INFO_USE_HTTP_AUTHENTICATION = 1 << 19
    RESPONSE_INFO_HAS_SIGNED_CERTIFICATE_TIMESTAMPS = 1 << 20
    RESPONSE_INFO_UNUSED_SINCE_PREFETCH = 1 << 21
    RESPONSE_INFO_HAS_KEY_EXCHANGE_GROUP = 1 << 22
    RESPONSE_INFO_PKP_BYPASSED = 1 << 23
    RESPONSE_INFO_HAS_STALENESS = 1 << 24
    RESPONSE_INFO_HAS_PEER_SIGNATURE_ALGORITHM = 1 << 25
    RESPONSE_INFO_RESTRICTED_PREFETCH = 1 << 26
    RESPONSE_INFO_HAS_DNS_ALIASES = 1 << 27
    RESPONSE_INFO_SINGLE_KEYED_CACHE_ENTRY_UNUSABLE = 1 << 28
    RESPONSE_INFO_ENCRYPTED_CLIENT_HELLO = 1 << 29
    RESPONSE_INFO_BROWSER_RUN_ID = 1 << 30
    RESPONSE_INFO_HAS_EXTRA_FLAGS = 1 << 31  # indicates that we need to read the extra flags after this value


class CachedMetadataExtraFlags(enum.IntFlag):
    RESPONSE_EXTRA_INFO_DID_USE_SHARED_DICTIONARY = 1
    RESPONSE_EXTRA_INFO_HAS_PROXY_CHAIN = 1 << 1
    RESPONSE_EXTRA_INFO_HAS_ORIGINAL_RESPONSE_TIME = 1 << 2


class CachedMetadata:
    # net/http/http_response_info.cc / net/http/http_response_info.h
    def __init__(
            self, header_declarations: set[str], header_attributes: dict[str, list[str]],
            request_time: datetime.datetime, response_time: datetime.datetime, certs: list[bytes],
            host_address: str, hot_port: int, other_attributes: dict[str, typing.Any]):
        self._declarations = header_declarations.copy()
        self._attributes = types.MappingProxyType(header_attributes.copy())
        self._request_time = request_time
        self._response_time = response_time
        self._certs = certs.copy()
        self._other_attributes = types.MappingProxyType(other_attributes)
        self._host_address = host_address
        self._host_port = hot_port

    @property
    def certs(self) -> typing.Iterable[bytes]:
        yield from self._certs

    @property
    def http_header_declarations(self) -> typing.Iterable[str]:
        yield from self._declarations

    @property
    def request_time(self) -> datetime.datetime:
        return self._request_time

    @property
    def response_time(self) -> datetime.datetime:
        return self._response_time

    @property
    def http_header_attributes(self) -> typing.Iterable[tuple[str, str]]:
        for key, vals in self._attributes.items():
            for val in vals:
                yield key, val

    def has_declaration(self, declaration: str) -> bool:
        return declaration in self._declarations

    def get_attribute(self, attribute: str) -> list[str]:
        return self._attributes.get(attribute.lower()) or []

    @property
    def other_cache_attributes(self):
        return self._other_attributes

    @classmethod
    def from_buffer(cls, buffer: bytes):
        # net/http/http_response_info.cc / net/http/http_response_info.h
        # and for the proxy chain:
        # net/base/proxy_chain.cc / net/base/proxy_server.h / net/base/proxy_server.cc
        # This is a pickle, but it's a very simple one so just align manually rather than use a pickle library
        # TODO: this is increasingly not "very simple", so we should move to using ccl_easy_chromium_pickle to tidy
        #  things up.
        reader = BinaryReader.from_bytes(buffer)
        total_length = reader.read_uint32()
        if total_length != len(buffer) - 4:
            raise ValueError("Metadata buffer is not the declared size")

        def align():
            alignment = reader.tell() % 4
            if alignment != 0:
                reader.read_raw(4 - alignment)

        flags = CachedMetadataFlags(reader.read_uint32())
        if flags & CachedMetadataFlags.RESPONSE_INFO_HAS_EXTRA_FLAGS:
            extra_flags = CachedMetadataExtraFlags(reader.read_uint32())
        else:
            extra_flags = CachedMetadataExtraFlags(0)

        request_time = reader.read_datetime()
        response_time = reader.read_datetime()

        if extra_flags & CachedMetadataExtraFlags.RESPONSE_EXTRA_INFO_HAS_ORIGINAL_RESPONSE_TIME:
            # not currently reported as the meaning is not clear, but needs to be read if present in the pickle
            original_response_time = reader.read_datetime()

        http_header_length = reader.read_uint32()
        http_header_raw = reader.read_raw(http_header_length)

        header_attributes: dict[str, list[str]] = {}
        header_declarations = set()

        for header_entry in http_header_raw.split(b"\00"):
            if not header_entry:
                continue  # skip empty entries
            parsed_entry = header_entry.decode("latin-1").split(":", 1)
            if len(parsed_entry) == 1:
                header_declarations.add(parsed_entry[0])
            elif len(parsed_entry) == 2:
                header_attributes.setdefault(parsed_entry[0].lower(), [])
                header_attributes[parsed_entry[0].lower()].append(parsed_entry[1].strip())

        other_attributes = {}

        certs = []
        if flags & CachedMetadataFlags.RESPONSE_INFO_HAS_CERT:
            # net/cert/x509_certificate.cc CreateFromPickle
            align()
            cert_count = reader.read_uint32()
            for _ in range(cert_count):
                align()
                cert_length = reader.read_uint32()
                certs.append(reader.read_raw(cert_length))

        if flags & CachedMetadataFlags.RESPONSE_INFO_HAS_CERT_STATUS:
            align()
            other_attributes["cert_status"] = reader.read_uint32()

        if flags & CachedMetadataFlags.RESPONSE_INFO_HAS_SECURITY_BITS:
            align()
            other_attributes["security_bits"] = reader.read_int32()

        if flags & CachedMetadataFlags.RESPONSE_INFO_HAS_SSL_CONNECTION_STATUS:
            align()
            other_attributes["ssl_connection_status"] = reader.read_int32()

        if flags & CachedMetadataFlags.RESPONSE_INFO_HAS_SIGNED_CERTIFICATE_TIMESTAMPS:
            align()
            # these are unused, only here for backwards compatibility
            ts_count = reader.read_int32()
            for _ in range(ts_count):
                # net/cert/signed_certificate_timestamp.cc
                ts_version = reader.read_int32()
                str_len = reader.read_int32()
                ts_log_id = reader.read_raw(str_len)
                align()
                ts_timestamp = reader.read_datetime()
                str_len = reader.read_int32()
                ts_extensions = reader.read_raw(str_len)
                align()
                ts_hash_algo = reader.read_int32()
                ts_sig_algo = reader.read_int32()
                str_len = reader.read_int32()
                ts_sig_data = reader.read_raw(str_len)
                align()
                ts_origin = reader.read_int32()
                str_len = reader.read_int32()
                ts_log_desc = reader.read_raw(str_len)
                align()
                ts_status = reader.read_uint16()
                align()

        if flags & CachedMetadataFlags.RESPONSE_INFO_HAS_VARY_DATA:
            # net/http/http_vary_data.cc InitFromPickle
            align()
            other_attributes["vary_data"] = reader.read_raw(16)

        host, port = None, None
        try:
            align()
            host_length = reader.read_uint32()
            host = reader.read_raw(host_length).decode("latin-1")
            align()
            port = reader.read_uint16()
        except ValueError:
            # bail out at this point if we've hit eof
            return cls(
                header_declarations, header_attributes, request_time, response_time, certs, host, port,
                other_attributes)

        # todo: there are other fields, they don't look too relevant for us in many cases,
        #  but I can return here to review.

        return cls(
            header_declarations, header_attributes, request_time, response_time, certs, host, port, other_attributes)


@dataclasses.dataclass(frozen=True)
class CacheFileLocation:
    file_name: str
    offset: int

    def __repr__(self):
        return f"<CacheFileLocation; file_name: '{self.file_name}'; offset: {self.offset}"

    def __str__(self):
        return f"{self.file_name} @ {self.offset}"


class ChromiumCache(abc.ABC):
    """
    Abstract base class that both forms of concrete cache types inherit from
    """
    def get_metadata(self, key: typing.Union[str, CacheKey]) -> list[typing.Optional[CachedMetadata]]:  # typing.Optional[CachedMetadata]:
        """
        :param key: the cache key for the entry
        :return: a list of CachedMetadata objects for this key. Most often this list will contain only one entry but
            this library can return old versions of records in some cases. The order of metadata should be the same as
            the records returned by get_cachefile
        """
        raise NotImplementedError()

    def get_cachefile(self, key: typing.Union[str, CacheKey]) -> list[bytes]:  # typing.Optional[bytes]:
        """
        :param key: the cache key for the entry
        :return: a list of bytes objects for this key containing the cached resource. Most often this list will contain
            only one entry but this library can return old versions of records in some cases. The order of data should
            be the same as the records returned by get_metadata
        """
        raise NotImplementedError()

    def get_location_for_metadata(self, key: typing.Union[str, CacheKey]) -> list[CacheFileLocation]:
        """
        :param key: the cache key for the entry
        :return: a list of CacheFileLocation objects for this key's metadata. Most often this list will contain only one
            entry but this library can return old versions of records in some cases. The order of metadata should be the
            same as the records returned by get_metadata
        """
        raise NotImplementedError()

    def get_location_for_cachefile(self, key: typing.Union[str, CacheKey]) -> list[CacheFileLocation]:
        """
        :param key: the cache key for the entry
        :return: a list of CacheFileLocation objects for this key's data. Most often this list will contain only one
            entry but this library can return old versions of records in some cases. The order of metadata should be the
            same as the records returned by get_metadata
        """
        raise NotImplementedError()

    def __enter__(self) -> "ChromiumCache":
        raise NotImplementedError()

    def __exit__(self, exc_type, exc_val, exc_tb):
        raise NotImplementedError()

    def keys(self) -> typing.Iterable[str]:
        """
        :return: yields the cache keys for this cache instance
        """
        raise NotImplementedError()

    def cache_keys(self) -> typing.Iterable[CacheKey]:
        """
        :return: yields the cache keys (as CacheKey objects) for this cache instance
        """
        raise NotImplementedError()


class ChromiumBlockFileCache(ChromiumCache):
    def __init__(self, cache_dir: typing.Union[os.PathLike, str]):
        self._in_dir = pathlib.Path(cache_dir)
        self._index_file = BlockFileIndexFile(self._in_dir / "index")
        self._block_files: dict[int, tuple[BlockFileHeader, typing.BinaryIO]] = {}
        self._keys = self._build_keys()

    def _get_block_file(self, block_file_number: int) -> tuple[BlockFileHeader, typing.BinaryIO]:
        if cached := self._block_files.get(block_file_number):
            return cached

        block_file_stream = (self._in_dir / f"data_{block_file_number}").open("rb")
        header = BlockFileHeader.from_bytes(block_file_stream.read(BlockFileHeader._BLOCK_HEADER_SIZE))
        self._block_files[block_file_number] = (header, block_file_stream)
        return header, block_file_stream

    def _build_keys(self):
        result = {}
        for addr in self._index_file.index:
            while addr.is_initialized:
                if not addr.sanity_check_for_entry():
                    print(f"Warning: Addr skipped as it is not sane for an entry: {addr}", file=sys.stderr)
                    break
                raw = self.get_data_for_addr(addr)
                try:
                    es = EntryStore.from_bytes(raw)
                except (ValueError, OverflowError):
                    print("Warning: EntryStore could not be read and it being skipped; bad data follows, if you "
                          "believe it to be a valid record, please contact the developer.", file=sys.stderr)
                    print(raw, file=sys.stderr)
                    break
                if es.key is not None:
                    key = es.key
                else:
                    key = self.get_data_for_addr(es.long_key_addr).decode("utf-8")[0:es.key_length]

                result[key] = es
                addr = es.next_entry

        return result

    def _get_location(self, key: str, stream_number: int):
        es = self._keys[key]
        addr = es.data_addrs[stream_number]
        if addr.file_type in _BLOCK_FILE_FILETYPE:
            file_name = f"data_{addr.file_selector}"
            block_header, stream = self._get_block_file(addr.file_selector)
            offset = BlockFileHeader._BLOCK_HEADER_SIZE + (block_header.entry_size * addr.block_number)
            return CacheFileLocation(file_name, offset)
        elif addr.file_type == FileType.EXTERNAL:
            file_name = f"f_{addr.external_file_number:06x}"
            return CacheFileLocation(file_name, 0)

        raise ValueError("unexpected file type")

    def get_location_for_metadata(self, key: typing.Union[str, CacheKey]) -> list[CacheFileLocation]:
        if isinstance(key, CacheKey):
            key = key.raw_key
        return [self._get_location(key, 0)]

    def get_location_for_cachefile(self, key: typing.Union[str, CacheKey]) -> list[CacheFileLocation]:
        if isinstance(key, CacheKey):
            key = key.raw_key
        return [self._get_location(key, 1)]

    def get_stream_for_addr(self, addr: Addr) -> typing.BinaryIO:
        if not addr.is_initialized:
            raise ValueError("Addr is not initialized")
        if addr.file_type in _BLOCK_FILE_FILETYPE:
            block_header, stream = self._get_block_file(addr.file_selector)
            stream.seek(BlockFileHeader._BLOCK_HEADER_SIZE + (block_header.entry_size * addr.block_number))
            return io.BytesIO(stream.read(block_header.entry_size * addr.contiguous_blocks))  # slow probably
        elif addr.file_type == FileType.EXTERNAL:
            return (self._in_dir / f"f_{addr.external_file_number:06x}").open("rb")

        raise ValueError("unexpected file type")

    def get_data_for_addr(self, addr: Addr) -> typing.Optional[bytes]:
        if not addr.is_initialized:
            raise ValueError("Addr is not initialized")
        if addr.file_type in _BLOCK_FILE_FILETYPE:
            block_header, stream = self._get_block_file(addr.file_selector)
            stream.seek(BlockFileHeader._BLOCK_HEADER_SIZE + (block_header.entry_size * addr.block_number))
            return stream.read(block_header.entry_size * addr.contiguous_blocks)
        elif addr.file_type == FileType.EXTERNAL:
            external_file_path = self._in_dir / f"f_{addr.external_file_number:06x}"
            if not external_file_path.exists():
                print(f"Warning: External cache file {external_file_path} is referenced in the data, but "
                      f"does not exist in the cache folder.", file=sys.stderr)
                return None
            with external_file_path.open("rb") as f:
                return f.read()

        raise ValueError("unexpected file type")

    def get_data_buffer(self, key: typing.Union[str, EntryStore, CacheKey], stream_number: int) -> typing.Optional[bytes]:
        if stream_number < 0 or stream_number > 2:
            raise ValueError("invalid stream number")
        if isinstance(key, EntryStore):
            es = key
        elif isinstance(key, CacheKey):
            es = self._keys[key.raw_key]
        else:
            es = self._keys[key]

        addr = es.data_addrs[stream_number]
        if not addr.is_initialized:
            return None

        data = self.get_data_for_addr(addr)
        if data is None:
            return None

        stream_length = es.data_sizes[stream_number]
        if data is not None and len(data) < stream_length:
            print(es, file=sys.stderr)
            raise ValueError(f"Could not get all of the data for stream {stream_number}")
        data = data[0:stream_length]
        return data

    def get_metadata(self, key: typing.Union[str, EntryStore, CacheKey]) -> list[typing.Optional[CachedMetadata]]:
        buffer = self.get_data_buffer(key, 0)
        if not buffer:
            return [None]
        meta = CachedMetadata.from_buffer(buffer)
        return [meta]

    def get_cachefile(self, key: typing.Union[str, EntryStore, CacheKey]) -> list[bytes]:
        return [self.get_data_buffer(key, 1)]

    def __enter__(self) -> "ChromiumBlockFileCache":
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def keys(self) -> typing.Iterable[str]:
        yield from self._keys.keys()

    def cache_keys(self) -> typing.Iterable[CacheKey]:
        for k in self._keys.keys():
            yield CacheKey(k)

    def values(self) -> typing.Iterable[EntryStore]:
        yield from self._keys.values()

    def items(self) -> typing.Iterable[tuple[str, EntryStore]]:
        yield from self._keys.items()

    def __contains__(self, item) -> bool:
        if isinstance(item, CacheKey):
            item = item.raw_key
        return item in self._keys

    def __getitem__(self, item) -> EntryStore:
        if isinstance(item, CacheKey):
            item = item.raw_key
        return self._keys[item]

    def close(self):
        for _, stream in self._block_files.values():
            stream.close()


@dataclasses.dataclass(frozen=True)
class SimpleCacheEOF:
    # net/disk_cache/simple/simple_entry_format.h
    flags: int
    data_crc: int
    stream_size: int

    _SIMPLE_FINAL_MAGIC: typing.ClassVar[int] = 0xf4fa6f45970d41d8  # is written little-endian in the file

    @classmethod
    def from_reader(cls, reader: BinaryReader):
        magic = reader.read_uint64()
        if magic != SimpleCacheEOF._SIMPLE_FINAL_MAGIC:
            raise ValueError(f"Invalid magic (expected {SimpleCacheEOF._SIMPLE_FINAL_MAGIC}; got {magic}")

        flags = reader.read_uint32()
        data_crc = reader.read_uint32()
        stream_size = reader.read_uint32()

        return cls(flags, data_crc, stream_size)

    @property
    def has_crc(self):
        return self.flags & 1 > 0

    @property
    def has_key_sha256(self):
        return self.flags & 2 > 0


@dataclasses.dataclass(frozen=True)
class SimpleCacheHeader:
    # net/disk_cache/simple/simple_entry_format.h
    version: int
    key_length: int
    key_hash: int

    _SIMPLE_INITIAL_MAGIC: typing.ClassVar[int] = 0xfcfb6d1ba7725c30  # is written little-endian in the file

    @classmethod
    def from_reader(cls, reader: BinaryReader):
        magic = reader.read_uint64()
        if magic != SimpleCacheHeader._SIMPLE_INITIAL_MAGIC:
            raise ValueError(f"Invalid magic (expected {SimpleCacheHeader._SIMPLE_INITIAL_MAGIC}; got {magic}")
        version = reader.read_uint32()
        key_length = reader.read_uint32()
        key_hash = reader.read_uint32()

        if EIGHT_BYTE_PICKLE_ALIGNMENT:
            _ = reader.read_uint32()  # need to align to 8 bytes before we get to the key

        return cls(version, key_length, key_hash)


class SimpleCacheFile:
    # net/disk_cache/simple/simple_entry_format.h

    def __init__(self, cache_file: typing.Union[os.PathLike, str]):
        self._path = pathlib.Path(cache_file)
        self._reader = BinaryReader(self._path.open("rb"))
        self._header = SimpleCacheHeader.from_reader(self._reader)
        self._key = self._reader.read_raw(self._header.key_length).decode("latin-1")

        # Peek forwards - are we at EOF? Sometimes (rarely) you only get a URL
        if self._reader.is_eof:
            self._stream_0_eof = None
            self._stream_1_eof = None
            self._stream_0_start_offset_negative = 0
            self._stream_1_start_offset = 0
            self._stream_1_length = 0

            self._has_data = False
            return
        else:
            self._has_data = True

        # get stream 0 EOF
        self._reader.seek(-SIMPLE_EOF_SIZE, os.SEEK_END)
        self._stream_0_eof = SimpleCacheEOF.from_reader(self._reader)
        self._stream_0_start_offset_negative = -SIMPLE_EOF_SIZE - self._stream_0_eof.stream_size
        if self._stream_0_eof.has_key_sha256:
            self._stream_0_start_offset_negative -= 32

        # get stream 1 EOF
        # the size of the stream 0 eof, the size of the stream 1 eof, the size of stream 0, 32 bytes if there's sha256
        self._reader.seek(-SIMPLE_EOF_SIZE - SIMPLE_EOF_SIZE - self._stream_0_eof.stream_size, os.SEEK_END)
        if self._stream_0_eof.has_key_sha256:
            self._reader.seek(-32, os.SEEK_CUR)
        stream_1_end_offset = self._reader.tell()
        # the eof for stream 1 might contain a stream length, but the comments in simple_entry_format.h say it won't?
        self._stream_1_eof = SimpleCacheEOF.from_reader(self._reader)
        self._stream_1_start_offset = SIMPLE_EOF_SIZE + self._header.key_length  # 20 = header length
        self._stream_1_length = stream_1_end_offset - self._stream_1_start_offset

    def __enter__(self) -> "SimpleCacheFile":
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def get_stream_0(self):
        if self._has_data:
            self._reader.seek(self._stream_0_start_offset_negative, os.SEEK_END)
            return self._reader.read_raw(self._stream_0_eof.stream_size)
        return b""

    def get_stream_1(self):
        if self._has_data:
            self._reader.seek(self._stream_1_start_offset, os.SEEK_SET)
            return self._reader.read_raw(self._stream_1_length)
        return b""

    @property
    def data_start_offset(self):
        return self._stream_1_start_offset

    @property
    def metadata_start_offset_negative(self):
        return self._stream_0_start_offset_negative

    @property
    def path(self) -> pathlib.Path:
        return self._path

    @property
    def key(self) -> str:
        return self._key

    @property
    def key_hash(self) -> int:
        return self._header.key_hash

    def close(self):
        self._reader.close()


class ChromiumSimpleFileCache(ChromiumCache):
    # net/disk_cache/simple/simple_entry_format.h
    _STREAM_0_1_FILENAME_PATTERN = re.compile(r"^[0-9a-f]{16}_0$")

    def __init__(self, cache_dir: typing.Union[os.PathLike, str]):
        self._cache_dir = pathlib.Path(cache_dir)
        self._file_lookup = types.MappingProxyType(self._build_keys())

    @property
    def cache_dir(self) -> pathlib.Path:
        return self._cache_dir

    def _build_keys(self) -> dict[str, list[pathlib.Path]]:
        # doing it this way is slow, but saves on having a million file handles open
        lookup: dict[str, list[pathlib.Path]] = {}
        for cache_file in self._cache_dir.iterdir():
            if cache_file.is_file() and ChromiumSimpleFileCache._STREAM_0_1_FILENAME_PATTERN.match(cache_file.name):
                with SimpleCacheFile(cache_file) as cf:
                    # if cf.key in lookup:
                    #     raise ValueError(f"{cf.key} already in lookup (please contact developer)")
                    lookup.setdefault(cf.key, [])
                    lookup[cf.key].append(cache_file)

        return lookup

    def get_location_for_metadata(self, key: typing.Union[str, CacheKey]) -> list[CacheFileLocation]:
        result = []
        if isinstance(key, CacheKey):
            key = key.raw_key
        for file in self._file_lookup[key]:
            file_length = file.stat().st_size
            with SimpleCacheFile(file) as cf:
                offset = file_length + cf.metadata_start_offset_negative
            result.append(CacheFileLocation(file.name, offset))
        return result

    def get_location_for_cachefile(self, key: typing.Union[str, CacheKey]) -> list[CacheFileLocation]:
        result = []
        if isinstance(key, CacheKey):
            key = key.raw_key
        for file in self._file_lookup[key]:
            with SimpleCacheFile(file) as cf:
                offset = cf.data_start_offset
            result.append(CacheFileLocation(file.name, offset))
        return result

    def get_metadata(self, key: typing.Union[str, CacheKey]) -> list[typing.Optional[CachedMetadata]]:
        result = []
        if isinstance(key, CacheKey):
            key = key.raw_key
        for file in self._file_lookup[key]:
            with SimpleCacheFile(file) as cf:
                buffer = cf.get_stream_0()
                if buffer:
                    result.append(CachedMetadata.from_buffer(buffer))
                else:
                    result.append(None)
        return result

    def get_cachefile(self, key: typing.Union[str, CacheKey]) -> list[bytes]:
        result = []
        if isinstance(key, CacheKey):
            key = key.raw_key
        for file in self._file_lookup[key]:
            with SimpleCacheFile(file) as cf:
                result.append(cf.get_stream_1())
        return result

    def __enter__(self) -> "ChromiumCache":
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def close(self):
        pass

    def keys(self) -> typing.Iterable[str]:
        yield from self._file_lookup.keys()

    def cache_keys(self) -> typing.Iterable[CacheKey]:
        for k in self._file_lookup.keys():
            yield CacheKey(k)

    def get_file_for_key(self, key: typing.Union[str, CacheKey]) -> list[str]:
        if isinstance(key, CacheKey):
            key = key.raw_key
        return [x.name for x in self._file_lookup[key]]


def guess_cache_class(
        cache_dir: typing.Optional[typing.Union[pathlib.Path, os.PathLike]]) \
        -> typing.Optional[typing.Type[typing.Union[ChromiumBlockFileCache, ChromiumSimpleFileCache]]]:
    cache_dir = pathlib.Path(cache_dir)
    data_files = {"data_0", "data_1", "data_2", "data_3"}

    for file in cache_dir.iterdir():
        # multiple tests to we can return as soon as possible
        if file.name == "index-dir":
            return ChromiumSimpleFileCache
        elif file.name in data_files:
            return ChromiumBlockFileCache
        elif re.match(r"f_[0-9a-f]{6}", file.name):
            return ChromiumBlockFileCache
        elif re.match(r"^[0-9a-f]{16}_0$", file.name):
            return ChromiumSimpleFileCache

    return None


def main(args):
    import csv
    import hashlib
    import mimetypes
    import brotli
    import gzip

    in_cache_dir = pathlib.Path(args[0])
    out_dir = pathlib.Path(args[1])
    cache_out_dir = out_dir / "cache_files"

    if not in_cache_dir.is_dir():
        raise ValueError("Input directory is not a directory or does not exist")

    if out_dir.exists():
        raise ValueError("Output directory already exists")

    out_dir.mkdir()
    cache_out_dir.mkdir()

    default_row_headers = ["file_hash", "key", "request_time", "response_time", "date"]
    dynamic_row_headers = set()
    rows: list[dict] = []

    cache_type = guess_cache_class(in_cache_dir)
    if cache_type is None:
        raise ValueError("Could not detect Chrome cache type")

    with cache_type(in_cache_dir) as cache:
        for key in cache.keys():
            out_extension = ""
            content_encoding = ""
            row = {"key": key}
            rows.append(row)

            metas = cache.get_metadata(key)
            datas = cache.get_cachefile(key)

            if len(metas) != len(datas):
                raise ValueError("Metadata records count does not match data records count")

            for meta, data in zip(metas, datas):
                if meta is not None:
                    row["request_time"] = meta.request_time
                    row["response_time"] = meta.response_time
                    for attribute, value in meta.http_header_attributes:
                        dynamic_row_headers.add(attribute)
                        if attribute in row:
                            row[attribute] += f"; {value}"
                        else:
                            row[attribute] = value

                    if mime := meta.get_attribute("content-type"):
                        out_extension = mimetypes.guess_extension(mime[0]) or ""

                    content_encoding = (meta.get_attribute("content-encoding") or [""])[0]

                #data = cache.get_cachefile(key)
                if data is not None:
                    if content_encoding.strip() == "gzip":
                        try:
                            data = gzip.decompress(data)
                        except (EOFError, gzip.BadGzipFile) as ex:
                            print(f"Warning: could not decompress data for key: \"{key}\"; reason: {ex}")
                    elif content_encoding.strip() == "br":
                        try:
                            data = brotli.decompress(data)
                        except brotli.error as ex:
                            print(f"Warning: could not decompress data for key: \"{key}\"; reason: {ex}")
                    elif content_encoding.strip() == "deflate":
                        try:
                            data = zlib.decompress(data, -zlib.MAX_WBITS)  # suppress trying to read a header
                        except zlib.error as ex:
                            print(f"Warning: could not decompress data for key: \"{key}\"; reason: {ex}")
                    elif content_encoding.strip() != "":
                        print(f"Warning: unknown content-encoding: {content_encoding}")

                    h = hashlib.sha256()
                    h.update(data)
                    cache_file_hash = h.hexdigest()
                    row["file_hash"] = cache_file_hash
                    with (cache_out_dir / (cache_file_hash + out_extension)).open("wb") as out:
                        out.write(data)
                else:
                    row["file_hash"] = "<No cache file data>"

    csv_out_f = (out_dir / "cache_report.csv").open("wt", encoding="utf-8", newline="")
    csv_out_f.write("\ufeff")
    csv_out = csv.DictWriter(
        csv_out_f, fieldnames=default_row_headers + sorted(dynamic_row_headers), dialect=csv.excel,
        quoting=csv.QUOTE_ALL, quotechar="\"", escapechar="\\")
    csv_out.writeheader()
    for row in rows:
        csv_out.writerow(row)

    csv_out_f.close()


if __name__ == '__main__':
    if len(sys.argv) < 3:
        print(f"USAGE: {pathlib.Path(sys.argv[0]).name} <cache input dir> <out dir>")
        exit(1)
    main(sys.argv[1:])



"""
Copyright 2020-2021, CCL Forensics

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import typing
import struct
import re
import os
import io
import pathlib
import dataclasses
import enum
from collections import namedtuple
from types import MappingProxyType

import ccl_simplesnappy

__version__ = "0.4"
__description__ = "A module for reading LevelDB databases"
__contact__ = "Alex Caithness"


def _read_le_varint(stream: typing.BinaryIO, *, is_google_32bit=False) -> typing.Optional[typing.Tuple[int, bytes]]:
    """Read varint from a stream.
    If the read is successful: returns a tuple of the (unsigned) value and the raw bytes making up that varint,
    otherwise returns None.
    Can be switched to limit the varint to 32 bit."""
    # this only outputs unsigned
    i = 0
    result = 0
    underlying_bytes = []
    limit = 5 if is_google_32bit else 10
    while i < limit:
        raw = stream.read(1)
        if len(raw) < 1:
            return None
        tmp, = raw
        underlying_bytes.append(tmp)
        result |= ((tmp & 0x7f) << (i * 7))
        if (tmp & 0x80) == 0:
            break
        i += 1
    return result, bytes(underlying_bytes)


def read_le_varint(stream: typing.BinaryIO, *, is_google_32bit=False) -> typing.Optional[int]:
    """Convenience version of _read_le_varint that only returns the value or None"""
    x = _read_le_varint(stream, is_google_32bit=is_google_32bit)
    if x is None:
        return None
    else:
        return x[0]


def read_length_prefixed_blob(stream: typing.BinaryIO) -> bytes:
    """Reads a blob of data which is prefixed with a varint length"""
    length = read_le_varint(stream)
    data = stream.read(length)
    if len(data) != length:
        raise ValueError(f"Could not read all data (expected {length}, got {len(data)}")
    return data


@dataclasses.dataclass(frozen=True)
class BlockHandle:
    """See: https://github.com/google/leveldb/blob/master/doc/table_format.md
    A BlockHandle contains an offset and length of a block in an ldb table file"""
    offset: int
    length: int

    @classmethod
    def from_stream(cls, stream: typing.BinaryIO) -> "BlockHandle":
        return cls(read_le_varint(stream), read_le_varint(stream))

    @classmethod
    def from_bytes(cls, data: bytes) -> "BlockHandle":
        with io.BytesIO(data) as stream:
            return BlockHandle.from_stream(stream)


@dataclasses.dataclass(frozen=True)
class RawBlockEntry:
    """Raw key, value for a record in a LDB file Block, along with the offset within the block from which it came from
    See: https://github.com/google/leveldb/blob/master/doc/table_format.md"""
    key: bytes
    value: bytes
    block_offset: int


class FileType(enum.Enum):
    Ldb = 1
    Log = 2


class KeyState(enum.Enum):
    Deleted = 0
    Live = 1
    Unknown = 2


@dataclasses.dataclass(frozen=True)
class Record:
    """A record from leveldb; includes details of the origin file, state, etc."""

    key: bytes
    value: bytes
    seq: int
    state: KeyState
    file_type: FileType
    origin_file: os.PathLike
    offset: int
    was_compressed: bool

    @property
    def user_key(self) -> bytes:
        """Returns the "userkey" which omits the metadata bytes which may reside at the end of the raw key"""
        if self.file_type == FileType.Ldb:
            if len(self.key) < 8:
                return self.key
            else:
                return self.key[0:-8]
        else:
            return self.key

    @classmethod
    def ldb_record(cls, key: bytes, value: bytes, origin_file: os.PathLike,
                   offset: int, was_compressed: bool) -> "Record":
        seq = (struct.unpack("<Q", key[-8:])[0]) >> 8
        if len(key) > 8:
            state = KeyState.Deleted if key[-8] == 0 else KeyState.Live
        else:
            state = KeyState.Unknown
        return cls(key, value, seq, state, FileType.Ldb, origin_file, offset, was_compressed)

    @classmethod
    def log_record(cls, key: bytes, value: bytes, seq: int, state: KeyState,
                   origin_file: os.PathLike, offset: int) -> "Record":
        return cls(key, value, seq, state, FileType.Log, origin_file, offset, False)


class Block:
    """Block from an .lldb (table) file. See: https://github.com/google/leveldb/blob/master/doc/table_format.md"""
    def __init__(self, raw: bytes, was_compressed: bool, origin: "LdbFile", offset: int):
        self._raw = raw
        self.was_compressed = was_compressed
        self.origin = origin
        self.offset = offset

        self._restart_array_count, = struct.unpack("<I", self._raw[-4:])
        self._restart_array_offset = len(self._raw) - (self._restart_array_count + 1) * 4

    def get_restart_offset(self, index) -> int:
        offset = self._restart_array_offset + (index * 4)
        return struct.unpack("<i", self._raw[offset: offset + 4])[0]

    def get_first_entry_offset(self) -> int:
        return self.get_restart_offset(0)

    def __iter__(self) -> typing.Iterable[RawBlockEntry]:
        offset = self.get_first_entry_offset()
        with io.BytesIO(self._raw) as buff:
            buff.seek(offset)

            key = b""

            while buff.tell() < self._restart_array_offset:
                start_offset = buff.tell()
                shared_length = read_le_varint(buff, is_google_32bit=True)
                non_shared_length = read_le_varint(buff, is_google_32bit=True)
                value_length = read_le_varint(buff, is_google_32bit=True)

                # sense check
                if offset >= self._restart_array_offset:
                    raise ValueError("Reading start of entry past the start of restart array")
                if shared_length > len(key):
                    raise ValueError("Shared key length is larger than the previous key")

                key = key[:shared_length] + buff.read(non_shared_length)
                value = buff.read(value_length)

                yield RawBlockEntry(key, value, start_offset)


class LdbFile:
    """A leveldb table (.ldb or .sst) file."""
    BLOCK_TRAILER_SIZE = 5
    FOOTER_SIZE = 48
    MAGIC = 0xdb4775248b80fb57

    def __init__(self, file: pathlib.Path):
        if not file.exists():
            raise FileNotFoundError(file)

        self.path = file
        self.file_no = int(file.stem, 16)

        self._f = file.open("rb")
        self._f.seek(-LdbFile.FOOTER_SIZE, os.SEEK_END)

        self._meta_index_handle = BlockHandle.from_stream(self._f)
        self._index_handle = BlockHandle.from_stream(self._f)
        self._f.seek(-8, os.SEEK_END)
        magic, = struct.unpack("<Q", self._f.read(8))
        if magic != LdbFile.MAGIC:
            raise ValueError(f"Invalid magic number in {file}")

        self._index = self._read_index()

    def _read_block(self, handle: BlockHandle) -> Block:
        # block is the size in the blockhandle plus the trailer
        # the trailer is 5 bytes long.
        # idx  size  meaning
        # 0    1     CompressionType (0 = none, 1 = snappy)
        # 1    4     CRC32

        self._f.seek(handle.offset)
        raw_block = self._f.read(handle.length)
        trailer = self._f.read(LdbFile.BLOCK_TRAILER_SIZE)

        if len(raw_block) != handle.length or len(trailer) != LdbFile.BLOCK_TRAILER_SIZE:
            raise ValueError(f"Could not read all of the block at offset {handle.offset} in file {self.path}")

        is_compressed = trailer[0] != 0
        if is_compressed:
            with io.BytesIO(raw_block) as buff:
                raw_block = ccl_simplesnappy.decompress(buff)

        return Block(raw_block, is_compressed, self, handle.offset)

    def _read_index(self) -> typing.Tuple[typing.Tuple[bytes, BlockHandle], ...]:
        index_block = self._read_block(self._index_handle)
        # key is earliest key, value is BlockHandle to that data block
        return tuple((entry.key, BlockHandle.from_bytes(entry.value))
                     for entry in index_block)

    def __iter__(self) -> typing.Iterable[Record]:
        """Iterate Records in this Table file"""
        for block_key, handle in self._index:
            block = self._read_block(handle)
            for entry in block:
                yield Record.ldb_record(
                    entry.key, entry.value, self.path,
                    block.offset if block.was_compressed else block.offset + entry.block_offset,
                    block.was_compressed)

    def close(self):
        self._f.close()


class LogEntryType(enum.IntEnum):
    Zero = 0
    Full = 1
    First = 2
    Middle = 3
    Last = 4


class LogFile:
    """A levelDb log (.log) file"""
    LOG_ENTRY_HEADER_SIZE = 7
    LOG_BLOCK_SIZE = 32768

    def __init__(self, file: pathlib.Path):
        if not file.exists():
            raise FileNotFoundError(file)

        self.path = file
        self.file_no = int(file.stem, 16)

        self._f = file.open("rb")

    def _get_raw_blocks(self) -> typing.Iterable[bytes]:
        self._f.seek(0)

        while chunk := self._f.read(LogFile.LOG_BLOCK_SIZE):
            yield chunk

    def _get_batches(self) -> typing.Iterable[typing.Tuple[int, bytes]]:
        in_record = False
        start_block_offset = 0
        block = b""
        for idx, chunk_ in enumerate(self._get_raw_blocks()):
            with io.BytesIO(chunk_) as buff:
                while buff.tell() < LogFile.LOG_BLOCK_SIZE - 6:
                    header = buff.read(7)
                    if len(header) < 7:
                        break
                    crc, length, block_type = struct.unpack("<IHB", header)

                    if block_type == LogEntryType.Full:
                        if in_record:
                            raise ValueError(f"Full block whilst still building a block at offset "
                                             f"{idx * LogFile.LOG_BLOCK_SIZE + buff.tell()} in {self.path}")
                        in_record = False
                        yield idx * LogFile.LOG_BLOCK_SIZE + buff.tell(), buff.read(length)
                    elif block_type == LogEntryType.First:
                        if in_record:
                            raise ValueError(f"First block whilst still building a block at offset "
                                             f"{idx * LogFile.LOG_BLOCK_SIZE + buff.tell()} in {self.path}")
                        start_block_offset = idx * LogFile.LOG_BLOCK_SIZE + buff.tell()
                        block = buff.read(length)
                        in_record = True
                    elif block_type == LogEntryType.Middle:
                        if not in_record:
                            raise ValueError(f"Middle block whilst not building a block at offset "
                                             f"{idx * LogFile.LOG_BLOCK_SIZE + buff.tell()} in {self.path}")
                        block += buff.read(length)
                    elif block_type == LogEntryType.Last:
                        if not in_record:
                            raise ValueError(f"Last block whilst not building a block at offset "
                                             f"{idx * LogFile.LOG_BLOCK_SIZE + buff.tell()} in {self.path}")
                        block += buff.read(length)
                        in_record = False
                        yield start_block_offset * LogFile.LOG_BLOCK_SIZE, block
                    else:
                        raise ValueError()  # Cannot happen

    def __iter__(self) -> typing.Iterable[Record]:
        """Iterate Records in this Log file"""
        for batch_offset, batch in self._get_batches():
            # as per write_batch and write_batch_internal
            # offset       length      description
            # 0            8           (u?)int64 Sequence number
            # 8            4           (u?)int32 Count - the log batch can contain multple entries
            #
            #         Then Count * the following:
            #
            # 12           1           ValueType (KeyState as far as this library is concerned)
            # 13           1-4         VarInt32 length of key
            # ...          ...         Key data
            # ...          1-4         VarInt32 length of value
            # ...          ...         Value data

            with io.BytesIO(batch) as buff:  # it's just easier this way
                header = buff.read(12)
                seq, count = struct.unpack("<QI", header)

                for i in range(count):
                    start_offset = batch_offset + buff.tell()
                    state = KeyState(buff.read(1)[0])
                    key_length = read_le_varint(buff, is_google_32bit=True)
                    key = buff.read(key_length)
                    # print(key)
                    if state != KeyState.Deleted:
                        value_length = read_le_varint(buff, is_google_32bit=True)
                        value = buff.read(value_length)
                    else:
                        value = b""

                    yield Record.log_record(key, value, seq + i, state, self.path, start_offset)

    def close(self):
        self._f.close()


class VersionEditTag(enum.IntEnum):
    """
    See: https://github.com/google/leveldb/blob/master/db/version_edit.cc
    """
    Comparator = 1,
    LogNumber = 2,
    NextFileNumber = 3,
    LastSequence = 4,
    CompactPointer = 5,
    DeletedFile = 6,
    NewFile = 7,
    # 8 was used for large value refs
    PrevLogNumber = 9


@dataclasses.dataclass(frozen=True)
class VersionEdit:
    """
    See:
    https://github.com/google/leveldb/blob/master/db/version_edit.h
    https://github.com/google/leveldb/blob/master/db/version_edit.cc
    """
    comparator: str = None
    log_number: int = None
    prev_log_number: int = None
    last_sequence: int = None
    next_file_number: int = None
    compaction_pointers: typing.Tuple[typing.Any] = tuple()
    deleted_files: typing.Tuple[typing.Any] = tuple()
    new_files: typing.Tuple[typing.Any] = tuple()

    @classmethod
    def from_buffer(cls, buffer: bytes):
        comparator = None
        log_number = None
        prev_log_number = None
        last_sequence = None
        next_file_number = None
        compaction_pointers = []
        deleted_files = []
        new_files = []

        compaction_pointer_nt = namedtuple("CompactionPointer", ["level", "pointer"])
        deleted_file_nt = namedtuple("DeletedFile", ["level", "file_no"])
        new_file_nt = namedtuple("NewFile", ["level", "file_no", "file_size", "smallest_key", "largest_key"])

        with io.BytesIO(buffer) as b:
            while b.tell() < len(buffer) - 1:
                tag = read_le_varint(b, is_google_32bit=True)

                if tag == VersionEditTag.Comparator:
                    comparator = read_length_prefixed_blob(b).decode("utf-8")
                elif tag == VersionEditTag.LogNumber:
                    log_number = read_le_varint(b)
                elif tag == VersionEditTag.PrevLogNumber:
                    prev_log_number = read_le_varint(b)
                elif tag == VersionEditTag.NextFileNumber:
                    next_file_number = read_le_varint(b)
                elif tag == VersionEditTag.LastSequence:
                    last_sequence = read_le_varint(b)
                elif tag == VersionEditTag.CompactPointer:
                    level = read_le_varint(b, is_google_32bit=True)
                    compaction_pointer = read_length_prefixed_blob(b)
                    compaction_pointers.append(compaction_pointer_nt(level, compaction_pointer))
                elif tag == VersionEditTag.DeletedFile:
                    level = read_le_varint(b, is_google_32bit=True)
                    file_no = read_le_varint(b)
                    deleted_files.append(deleted_file_nt(level, file_no))
                elif tag == VersionEditTag.NewFile:
                    level = read_le_varint(b, is_google_32bit=True)
                    file_no = read_le_varint(b)
                    file_size = read_le_varint(b)
                    smallest = read_length_prefixed_blob(b)
                    largest = read_length_prefixed_blob(b)
                    new_files.append(new_file_nt(level, file_no, file_size, smallest, largest))

        return cls(comparator, log_number, prev_log_number, last_sequence, next_file_number, tuple(compaction_pointers),
            tuple(deleted_files), tuple(new_files))


class ManifestFile:
    """
    Represents a manifest file which contains database metadata.
    Manifest files are, at a high level, formatted like a log file in terms of the block and batch format,
    but the data within the batches follow their own format.

    Main use is to identify the level of files, use `file_to_level` property to look up levels based on file no.

    See:
    https://github.com/google/leveldb/blob/master/db/version_edit.h
    https://github.com/google/leveldb/blob/master/db/version_edit.cc
    """

    MANIFEST_FILENAME_PATTERN = "MANIFEST-([0-9A-F]{6})"

    def __init__(self, path: pathlib.Path):
        if match := re.match(ManifestFile.MANIFEST_FILENAME_PATTERN, path.name):
            self.file_no = int(match.group(1))
        else:
            raise ValueError("Invalid name for Manifest")

        self._f = path.open("rb")
        self.path = path

        self.file_to_level = {}
        for edit in self:
            if edit.new_files:
                for nf in edit.new_files:
                    self.file_to_level[nf.file_no] = nf.level

        self.file_to_level = MappingProxyType(self.file_to_level)

    def _get_raw_blocks(self) -> typing.Iterable[bytes]:
        self._f.seek(0)

        while chunk := self._f.read(LogFile.LOG_BLOCK_SIZE):
            yield chunk

    def _get_batches(self) -> typing.Iterable[typing.Tuple[int, bytes]]:
        in_record = False
        start_block_offset = 0
        block = b""
        for idx, chunk_ in enumerate(self._get_raw_blocks()):
            with io.BytesIO(chunk_) as buff:
                while buff.tell() < LogFile.LOG_BLOCK_SIZE - 6:
                    header = buff.read(7)
                    if len(header) < 7:
                        break
                    crc, length, block_type = struct.unpack("<IHB", header)

                    if block_type == LogEntryType.Full:
                        if in_record:
                            raise ValueError(f"Full block whilst still building a block at offset "
                                             f"{idx * LogFile.LOG_BLOCK_SIZE + buff.tell()} in {self.path}")
                        in_record = False
                        yield idx * LogFile.LOG_BLOCK_SIZE + buff.tell(), buff.read(length)
                    elif block_type == LogEntryType.First:
                        if in_record:
                            raise ValueError(f"First block whilst still building a block at offset "
                                             f"{idx * LogFile.LOG_BLOCK_SIZE + buff.tell()} in {self.path}")
                        start_block_offset = idx * LogFile.LOG_BLOCK_SIZE + buff.tell()
                        block = buff.read(length)
                        in_record = True
                    elif block_type == LogEntryType.Middle:
                        if not in_record:
                            raise ValueError(f"Middle block whilst not building a block at offset "
                                             f"{idx * LogFile.LOG_BLOCK_SIZE + buff.tell()} in {self.path}")
                        block += buff.read(length)
                    elif block_type == LogEntryType.Last:
                        if not in_record:
                            raise ValueError(f"Last block whilst not building a block at offset "
                                             f"{idx * LogFile.LOG_BLOCK_SIZE + buff.tell()} in {self.path}")
                        block += buff.read(length)
                        in_record = False
                        yield start_block_offset * LogFile.LOG_BLOCK_SIZE, block
                    else:
                        raise ValueError()  # Cannot happen

    def __iter__(self):
        for batch_offset, batch in self._get_batches():
            yield VersionEdit.from_buffer(batch)

    def close(self):
        self._f.close()


class RawLevelDb:
    DATA_FILE_PATTERN = r"[0-9]{6}\.(ldb|log|sst)"

    def __init__(self, in_dir: os.PathLike):

        self._in_dir = pathlib.Path(in_dir)
        if not self._in_dir.is_dir():
            raise ValueError("in_dir is not a directory")

        self._files = []
        latest_manifest = (0, None)
        for file in self._in_dir.iterdir():
            if file.is_file() and re.match(RawLevelDb.DATA_FILE_PATTERN, file.name):
                if file.suffix.lower() == ".log":
                    self._files.append(LogFile(file))
                elif file.suffix.lower() == ".ldb" or file.suffix.lower() == ".sst":
                    self._files.append(LdbFile(file))
            if file.is_file() and re.match(ManifestFile.MANIFEST_FILENAME_PATTERN, file.name):
                manifest_no = int(re.match(ManifestFile.MANIFEST_FILENAME_PATTERN, file.name).group(1), 16)
                if latest_manifest[0] < manifest_no:
                    latest_manifest = (manifest_no, file)

        self.manifest = ManifestFile(latest_manifest[1]) if latest_manifest[1] is not None else None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    @property
    def in_dir_path(self) -> pathlib.Path:
        return self._in_dir

    def iterate_records_raw(self, *, reverse=False) -> typing.Iterable[Record]:
        for file_containing_records in sorted(self._files, reverse=reverse, key=lambda x: x.file_no):
            yield from file_containing_records

    def close(self):
        for file in self._files:
            file.close()
        if self.manifest:
            self.manifest.close()



"""
Copyright 2020, CCL Forensics

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import sys
import struct
import datetime
import types
import typing
import re

__version__ = "0.1.1"
__description__ = "Partial reimplementation of the V8 Javascript Object Serialization"
__contact__ = "Alex Caithness"

# TODO: We need to address cyclic references, which are permissible. Probably take the same approach as in ccl_bplist
#  and subclass the collection types to resolve references JIT

# See: https://github.com/v8/v8/blob/master/src/objects/value-serializer.cc

__DEBUG = False


def log(msg, debug_only=True):
    if not debug_only or __DEBUG:
        caller_name = sys._getframe(1).f_code.co_name
        caller_line = sys._getframe(1).f_code.co_firstlineno
        print(f"{caller_name} ({caller_line}):\t{msg}")


def read_le_varint(stream: typing.BinaryIO, is_32bit=False) -> typing.Optional[typing.Tuple[int, bytes]]:
    # this only outputs unsigned
    limit = 5 if is_32bit else 10
    i = 0
    result = 0
    underlying_bytes = []
    while i < limit:  # 64 bit max possible?
        raw = stream.read(1)
        if len(raw) < 1:
            return None
        tmp, = raw
        underlying_bytes.append(tmp)
        result |= ((tmp & 0x7f) << (i * 7))
        if (tmp & 0x80) == 0:
            break
        i += 1
    return result, bytes(underlying_bytes)


class _Undefined:
    def __bool__(self):
        return False

    def __eq__(self, other):
        if isinstance(other, _Undefined):
            return True
        return False

    def __repr__(self):
        return "<Undefined>"

    def __str__(self):
        return "<Undefined>"


class SharedObject:
    def __init__(self, object_id: int):
        self.id = object_id

    def __repr__(self):
        return f"<SharedObject; id: {self.id}>"

    def __str__(self):
        return f"<SharedObject; id: {self.id}>"


class Constants:
    # Constants
    kLatestVersion = 15

    # version:uint32_t (if at beginning of data, sets version > 0)
    token_kVersion = b"\xFF"
    # ignore
    token_kPadding = b"\0"
    # refTableSize:uint32_t (previously used for sanity checks; safe to ignore)
    token_kVerifyObjectCount = b"?"
    # Oddballs (no data).
    token_kTheHole = b"-"
    token_kUndefined = b"_"
    token_kNull = b"0"
    token_kTrue = b"T"
    token_kFalse = b"F"
    # Number represented as 32-bit integer, ZigZag-encoded
    # (like sint32 in protobuf)
    token_kInt32 = b"I"
    # Number represented as 32-bit unsigned integer, varint-encoded
    # (like uint32 in protobuf)
    token_kUint32 = b"U"
    # Number represented as a 64-bit double.
    # Host byte order is used (N.B. this makes the format non-portable).
    token_kDouble = b"N"
    # BigInt. Bitfield:uint32_t, then raw digits storage.
    token_kBigInt = b"Z"
    # byteLength:uint32_t, then raw data
    token_kUtf8String = b"S"
    token_kOneByteString = b"\""
    token_kTwoByteString = b"c"
    # Reference to a serialized object. objectID:uint32_t
    token_kObjectReference = b"^"
    # Beginning of a JS object.
    token_kBeginJSObject = b"o"
    # End of a JS object. numProperties:uint32_t
    token_kEndJSObject = b"{"
    # Beginning of a sparse JS array. length:uint32_t
    # Elements and properties are written as token_key/value pairs, like objects.
    token_kBeginSparseJSArray = b"a"
    # End of a sparse JS array. numProperties:uint32_t length:uint32_t
    token_kEndSparseJSArray = b"@"
    # Beginning of a dense JS array. length:uint32_t
    # |length| elements, followed by properties as token_key/value pairs
    token_kBeginDenseJSArray = b"A"
    # End of a dense JS array. numProperties:uint32_t length:uint32_t
    token_kEndDenseJSArray = b"$"
    # Date. millisSinceEpoch:double
    token_kDate = b"D"
    # Boolean object. No data.
    token_kTrueObject = b"y"
    token_kFalseObject = b"x"
    # Number object. value:double
    token_kNumberObject = b"n"
    # BigInt object. Bitfield:uint32_t, then raw digits storage.
    token_kBigIntObject = b"z"
    # String object, UTF-8 encoding. byteLength:uint32_t, then raw data.
    token_kStringObject = b"s"
    # Regular expression, UTF-8 encoding. byteLength:uint32_t, raw data
    # flags:uint32_t.
    token_kRegExp = b"R"
    # Beginning of a JS map.
    token_kBeginJSMap = b";"
    # End of a JS map. length:uint32_t.
    token_kEndJSMap = b":"
    # Beginning of a JS set.
    token_kBeginJSSet = b"'"
    # End of a JS set. length:uint32_t.
    token_kEndJSSet = b","
    # Array buffer. byteLength:uint32_t, then raw data.
    token_kArrayBuffer = b"B"
    # Array buffer (transferred). transferID:uint32_t
    token_kArrayBufferTransfer = b"t"
    # View into an array buffer.
    # subtag:ArrayBufferViewTag, byteOffset:uint32_t, byteLength:uint32_t
    # For typed arrays, byteOffset and byteLength must be divisible by the size
    # of the element.
    # Note: token_kArrayBufferView is special, and should have an ArrayBuffer (or an
    # ObjectReference to one) serialized just before it. This is a quirk arising
    # from the previous stack-based implementation.
    token_kArrayBufferView = b"V"
    # Shared array buffer. transferID:uint32_t
    token_kSharedArrayBuffer = b"u"
    # A HeapObject shared across Isolates.sharedValueID: uint32_t
    token_kSharedObject = 'p'
    # A wasm module object transfer. next value is its index.
    token_kWasmModuleTransfer = b"w"
    # The delegate is responsible for processing all following data.
    # This "escapes" to whatever wire format the delegate chooses.
    token_kHostObject = b"\\"
    # A transferred WebAssembly.Memory object. maximumPages:int32_t, then by
    # SharedArrayBuffer tag and its data.
    token_kWasmMemoryTransfer = b"m"
    # A list of (subtag: ErrorTag, [subtag dependent data]). See ErrorTag for
    # details.
    token_kError = b"r"

    # The following tags are reserved because they were in use in Chromium before
    # the token_kHostObject tag was introduced in format version 13, at
    #   v8           refs/heads/master@{#43466}
    #   chromium/src refs/heads/master@{#453568}
    #
    # They must not be reused without a version check to prevent old values from
    # starting to deserialize incorrectly. For simplicity, it's recommended to
    # avoid them altogether.
    #
    # This is the set of tags that existed in SerializationTag.h at that time and
    # still exist at the time of this writing (i.e., excluding those that were
    # removed on the Chromium side because there should be no real user data
    # containing them).
    #
    # It might be possible to also free up other tags which were never persisted
    # (e.g. because they were used only for transfer) in the future.
    token_kLegacyReservedMessagePort = b"M"
    token_kLegacyReservedBlob = b"b"
    token_kLegacyReservedBlobIndex = b"i"
    token_kLegacyReservedFile = b"f"
    token_kLegacyReservedFileIndex = b"e"
    token_kLegacyReservedDOMFileSystem = b"d"
    token_kLegacyReservedFileList = b"l"
    token_kLegacyReservedFileListIndex = b"L"
    token_kLegacyReservedImageData = b"#"
    token_kLegacyReservedImageBitmap = b"g"
    token_kLegacyReservedImageBitmapTransfer = b"G"
    token_kLegacyReservedOffscreenCanvas = b"H"
    token_kLegacyReservedCryptoKey = b"token_k"
    token_kLegacyReservedRTCCertificate = b"token_k"


class ArrayBufferViewTag:
    tag_kInt8Array = "b"
    tag_kUint8Array = "B"
    tag_kUint8ClampedArray = "C"
    tag_kInt16Array = "w"
    tag_kUint16Array = "W"
    tag_kInt32Array = "d"
    tag_kUint32Array = "D"
    tag_kFloat32Array = "f"
    tag_kFloat64Array = "F"
    tag_kBigInt64Array = "q"
    tag_kBigUint64Array = "Q"
    tag_kDataView = "?"

    STRUCT_LOOKUP = types.MappingProxyType({
        tag_kInt8Array: "b",
        tag_kUint8Array: "B",
        tag_kUint8ClampedArray: "B",
        tag_kInt16Array: "h",
        tag_kUint16Array: "H",
        tag_kInt32Array: "i",
        tag_kUint32Array: "I",
        tag_kFloat32Array: "f",
        tag_kFloat64Array: "d",
        tag_kBigInt64Array: "q",
        tag_kBigUint64Array: "Q",
        tag_kDataView: "c"
    })


class Deserializer:
    Undefined = _Undefined()

    __ODDBALLS = {
        Constants.token_kUndefined: Undefined,
        Constants.token_kTheHole: Undefined,
        Constants.token_kNull: None,
        Constants.token_kTrue: True,
        Constants.token_kFalse: False,
    }

    __WRAPPED_PRIMITIVES = {
        Constants.token_kTrueObject,
        Constants.token_kFalseObject,
        Constants.token_kNumberObject,
        Constants.token_kBigIntObject,
        Constants.token_kStringObject
    }

    def __init__(self, stream: typing.BinaryIO, host_object_delegate: typing.Callable,
                 *, is_little_endian=True, is_64bit=True):
        self._f = stream
        self._host_object_delegate = host_object_delegate
        self._endian = "<" if is_little_endian else ">"
        self._pointer_size = 8 if is_64bit else 4
        self._next_id = 0
        self._objects = []
        self.version = self._read_header()

    def _read_raw(self, length: int) -> bytes:
        start = self._f.tell()
        raw = self._f.read(length)
        if len(raw) != length:
            raise ValueError(f"Could not read all data at offset {start}; wanted {length}; got {len(raw)}")

        return raw

    def _read_le_varint(self) -> typing.Optional[typing.Tuple[int, bytes]]:
        return read_le_varint(self._f)

    def _read_zigzag(self) -> int:
        unsigned = self._read_le_varint()[0]
        if unsigned & 1:
            return -(unsigned >> 1)
        else:
            return unsigned >> 1

    def _read_unit32(self) -> int:
        return self._read_le_varint()[0]

    def _read_double(self) -> float:
        return struct.unpack(f"{self._endian}d", self._read_raw(8))[0]

    # def _read_uint32(self) -> int:
    #     return self._read_le_varint()

    # def _read_uint64(self) -> int:
    #     return self._read_le_varint()

    def _read_bigint(self) -> int:
        size_flag = self._read_le_varint()[0]
        is_neg = size_flag & 0x01
        size = size_flag >> 4
        raw = self._read_raw(size * self._pointer_size)

        value = int.from_bytes(raw, "big" if self._endian == ">" else "little", signed=False)
        if is_neg:
            value = -value

        return value

    def _read_utf8_string(self) -> str:
        length = self._read_le_varint()[0]
        return self._read_raw(length).decode("utf8")

    def _read_one_byte_string(self) -> typing.AnyStr:
        length = self._read_le_varint()[0]
        # I think this can be used to store raw 8-bit data, so return ascii if we can, otherwise bytes
        raw = self._read_raw(length)  # .decode("ascii")
        try:
            result = raw.decode("ascii")
        except UnicodeDecodeError:
            result = raw
        return result

    def _read_two_byte_string(self) -> str:
        length = self._read_le_varint()[0]
        return self._read_raw(length).decode("utf-16-le")  # le?

    def _read_string(self) -> str:
        if self.version < 12:
            return self._read_utf8_string()

        value = self._read_object()
        assert isinstance(value, str)

        return value

    def _read_object_by_reference(self) -> typing.Any:
        ref_id = self._read_le_varint()[0]
        return self._objects[ref_id]

    def _read_tag(self) -> bytes:
        while True:
            t = self._f.read(1)
            if t != Constants.token_kPadding:
                return t

    def _peek_tag(self) -> bytes:
        start = self._f.tell()
        tag = self._read_tag()
        self._f.seek(start, 0)
        return tag

    def _read_date(self) -> datetime.datetime:
        x = self._read_double()
        result = datetime.datetime(1970, 1, 1) + datetime.timedelta(milliseconds=x)
        self._objects.append(result)
        return result

    def _read_js_regex(self) -> typing.Pattern:
        log(f"Reading js regex properties at {self._f.tell()}")
        pattern = self._read_string()
        flags = self._read_le_varint()

        # TODO: Flags?
        regex = re.compile(pattern)
        self._objects.append(regex)
        return regex

    def _read_js_object_properties(self, end_tag) -> typing.Iterable[typing.Tuple[typing.Any, typing.Any]]:
        log(f"Reading object properties at {self._f.tell()} with end tag: {end_tag}")
        while True:
            if self._peek_tag() == end_tag:
                log(f"Object end at offset {self._f.tell()}")
                break
            key = self._read_object()
            value = self._read_object()

            yield key, value

        assert self._read_tag() == end_tag

    def _read_js_object(self) -> dict:
        log(f"Reading js object properties at {self._f.tell()}")
        result = {}
        self._objects.append(result)
        for key, value in self._read_js_object_properties(Constants.token_kEndJSObject):
            result[key] = value
        # while True:
        #     if self._peek_tag() == end_tag:
        #         log(f"Object end at offset {self._f.tell()}")
        #         break
        #     key = self._read_object()
        #     value = self._read_object()
        #     result[key] = value
        #
        # assert self._read_tag() == end_tag
        property_count = self._read_le_varint()[0]
        log(f"Actual property count: {len(result)}; stated property count: {property_count}")
        if len(result) != property_count:
            raise ValueError("Property count mismatch")

        return result

    def _read_js_sparse_array(self) -> list:
        log(f"Reading js sparse array properties at {self._f.tell()}")
        # TODO: implement a sparse list so that this isn't so horribly inefficient
        length = self._read_le_varint()[0]
        result = [None for _ in range(length)]
        self._objects.append(result)

        sparse_object = self._read_js_object_properties(Constants.token_kEndSparseJSArray)
        prop_count = 0
        for key, value in sparse_object:
            i = int(key)
            result[i] = value
            prop_count += 1
        expected_num_properties = self._read_le_varint()[0]

        log(f"Actual property count: {prop_count}; stated property count: {expected_num_properties}")
        if prop_count != expected_num_properties:
            raise ValueError("Property count mismatch")

        expected_length = self._read_le_varint()[0]  # TODO: should this be checked?

        return result

    def _read_js_dense_array(self) -> list:
        log(f"Reading js dense array properties at {self._f.tell()}")
        length = self._read_le_varint()[0]
        result = [None for _ in range(length)]
        self._objects.append(result)

        for i in range(length):
            result[i] = self._read_object()

        # And then there's a sparse bit maybe?
        sparse_object = self._read_js_object_properties(Constants.token_kEndDenseJSArray)
        prop_count = 0
        for key, value in sparse_object:
            i = int(key)
            result[i] = value
            prop_count += 1

        expected_num_properties = self._read_le_varint()[0]

        log(f"Actual property count: {prop_count}; stated property count: {expected_num_properties}")
        if prop_count != expected_num_properties:
            raise ValueError("Property count mismatch")

        expected_length = self._read_le_varint()[0]  # TODO: should this be checked?

        return result

    def _read_js_map(self) -> dict:
        log(f"Reading js map at {self._f.tell()}")
        result = {}
        self._objects.append(result)
        while True:
            if self._peek_tag() == Constants.token_kEndJSMap:
                log(f"End of map at {self._f.tell()}")
                break

            key = self._read_object()
            value = self._read_object()
            result[key] = value

        assert self._read_tag() == Constants.token_kEndJSMap

        expected_length = self._read_le_varint()[0]
        log(f"Actual map item count: {len(result) * 2}; stated map item count: {expected_length}")
        if expected_length != len(result) * 2:
            raise ValueError("Map count mismatch")

        return result

    def _read_js_set(self) -> set:
        log(f"Reading js set properties at {self._f.tell()}")
        result = set()
        self._objects.append(result)

        while True:
            if self._peek_tag() == Constants.token_kEndJSSet:
                log(f"End of set at {self._f.tell()}")
                break

            result.add(self._read_object())

        assert self._read_tag() == Constants.token_kEndJSSet

        expected_length = self._read_le_varint()[0]
        log(f"Actual set item count: {len(result)}; stated set item count: {expected_length}")
        if expected_length != len(result):
            raise ValueError("Set count mismatch")

        return result

    def _read_js_arraybuffer(self) -> bytes:
        length = self._read_le_varint()[0]
        raw = self._read_raw(length)
        self._objects.append(raw)

        return raw

    def _wrap_js_array_buffer_view(self, raw: bytes) -> tuple:
        if not isinstance(raw, bytes):
            raise TypeError("Only bytes should be passed to be wrapped in a buffer view")

        log(f"Wrapping in ArrayBufferView at offset {self._f.tell()}")

        tag = chr(self._read_le_varint()[0])
        byte_offset = self._read_le_varint()[0]
        byte_length = self._read_le_varint()[0]

        if byte_offset + byte_length > len(raw):
            raise ValueError("Not enough data in the raw data to hold the defined data")

        # See: https://github.com/v8/v8/blob/4d34ea98bb655295ab1f9003f6783bd509b7ccb3/src/objects/value-serializer.cc#L1967
        if self.version >= 14:
            flags = self._read_le_varint()[0]

        log(f"ArrayBufferView: tag: {tag}; byte_offset: {byte_offset}; byte_length: {byte_length}")

        fmt = ArrayBufferViewTag.STRUCT_LOOKUP[tag]
        element_length = struct.calcsize(fmt)
        if byte_length % element_length != 0:
            raise ValueError(f"ArrayBufferView doesn't fit nicely: byte_length: {byte_length}; "
                             f"element_length: {element_length}")

        element_count = byte_length // element_length

        return struct.unpack(f"{self._endian}{element_count}{fmt}", raw[byte_offset: byte_offset + byte_length])

    def _read_host_object(self) -> typing.Any:
        result = self._host_object_delegate(self._f)
        self._objects.append(result)
        return result

    def _read_shared_object(self) -> SharedObject:
        shobj_id = self._read_le_varint()[0]
        return SharedObject(shobj_id)

    def _not_implemented(self):
        raise NotImplementedError("Todo")

    def _read_object_internal(self) -> typing.Tuple[bytes, typing.Any]:
        tag = self._read_tag()

        log(f"Offset: {self._f.tell()}; Tag: {tag}")

        if tag in Deserializer.__ODDBALLS:
            return tag, Deserializer.__ODDBALLS[tag]

        func = {
            Constants.token_kTrueObject: lambda: Deserializer.__ODDBALLS[Constants.token_kTrue],
            Constants.token_kFalseObject: lambda: Deserializer.__ODDBALLS[Constants.token_kFalse],
            Constants.token_kNumberObject: self._read_double,
            Constants.token_kUint32: self._read_unit32,
            Constants.token_kInt32: self._read_zigzag,
            Constants.token_kDouble: self._read_double,
            Constants.token_kDate: self._read_date,
            Constants.token_kBigInt: self._read_bigint,
            Constants.token_kBigIntObject: self._read_bigint,
            Constants.token_kUtf8String: self._read_utf8_string,
            Constants.token_kOneByteString: self._read_one_byte_string,
            Constants.token_kTwoByteString: self._read_two_byte_string,
            Constants.token_kStringObject: self._read_string,
            Constants.token_kRegExp: self._read_js_regex,
            Constants.token_kObjectReference: self._read_object_by_reference,
            Constants.token_kBeginJSObject: self._read_js_object,
            Constants.token_kSharedObject: self._read_shared_object,
            Constants.token_kBeginSparseJSArray: self._read_js_sparse_array,
            Constants.token_kBeginDenseJSArray: self._read_js_dense_array,
            Constants.token_kBeginJSMap: self._read_js_map,
            Constants.token_kBeginJSSet: self._read_js_set,
            Constants.token_kArrayBuffer: self._read_js_arraybuffer,
            Constants.token_kSharedArrayBuffer: self._not_implemented,  # and probably never, as it can't be pulled from the data I think?
            Constants.token_kArrayBufferTransfer: self._not_implemented,
            Constants.token_kError: self._not_implemented,
            Constants.token_kWasmModuleTransfer: self._not_implemented,
            Constants.token_kWasmMemoryTransfer: self._not_implemented,
            Constants.token_kHostObject: self._read_host_object,
        }.get(tag)

        if func is None:
            raise ValueError(f"Unknown tag {tag}")

        value = func()

        if tag in Deserializer.__WRAPPED_PRIMITIVES:
            self._objects.append(value)

        return tag, value

    def _read_object(self) -> typing.Any:
        log(f"Read object at offset: {self._f.tell()}")
        tag, o = self._read_object_internal()

        if self._peek_tag() == Constants.token_kArrayBufferView:
            assert self._read_tag() == Constants.token_kArrayBufferView
            o = self._wrap_js_array_buffer_view(o)

        return o

    def _read_header(self) -> int:
        tag = self._read_tag()
        if tag != Constants.token_kVersion:
            raise ValueError("Didn't get version tag in the header")
        version = self._read_le_varint()[0]
        return version

    def read(self) -> typing.Any:
        return self._read_object()



"""
Copyright 2021-2024, CCL Forensics
Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import io
import bisect
import re
import sys
import pathlib
import types
import typing
import collections.abc as col_abc
import dataclasses
import datetime

from .storage_formats import ccl_leveldb
from .common import KeySearch

__version__ = "0.5"
__description__ = "Module for reading the Chromium leveldb localstorage format"
__contact__ = "Alex Caithness"

"""
See: https://source.chromium.org/chromium/chromium/src/+/main:components/services/storage/dom_storage/local_storage_impl.cc
Meta keys:
    Key = "META:" + storage_key (the host)
    Value = protobuff: 1=timestamp (varint); 2=size in bytes (varint)
    
Record keys:
    Key = "_" + storage_key + "\\x0" + script_key
    Value = record_value
    
"""

_META_PREFIX = b"META:"
_RECORD_KEY_PREFIX = b"_"
_CHROME_EPOCH = datetime.datetime(1601, 1, 1, 0, 0, 0)

EIGHT_BIT_ENCODING = "iso-8859-1"


def from_chrome_timestamp(microseconds: int) -> datetime.datetime:
    return _CHROME_EPOCH + datetime.timedelta(microseconds=microseconds)


def decode_string(raw: bytes) -> str:
    """
    decodes a type-prefixed string - prefix of: 0=utf-16-le; 1=an extended ascii codepage (likely dependant on locale)
    :param raw: raw prefixed-string data
    :return: decoded string
    """
    prefix = raw[0]
    if prefix == 0:
        return raw[1:].decode("utf-16-le")
    elif prefix == 1:
        return raw[1:].decode(EIGHT_BIT_ENCODING)
    else:
        raise ValueError("Unexpected prefix, please contact developer")


@dataclasses.dataclass(frozen=True)
class StorageMetadata:
    storage_key: str
    timestamp: datetime.datetime
    size_in_bytes: int
    leveldb_seq_number: int

    @classmethod
    def from_protobuff(cls, storage_key: str, data: bytes, seq: int):
        with io.BytesIO(data) as stream:
            # This is a simple protobuff, so we'll read it directly, but with checks, rather than add a dependency
            ts_tag = ccl_leveldb.read_le_varint(stream)
            if (ts_tag & 0x07) != 0 or (ts_tag >> 3) != 1:
                raise ValueError("Unexpected tag when reading StorageMetadata from protobuff")
            timestamp = from_chrome_timestamp(ccl_leveldb.read_le_varint(stream))

            size_tag = ccl_leveldb.read_le_varint(stream)
            if (size_tag & 0x07) != 0 or (size_tag >> 3) != 2:
                raise ValueError("Unexpected tag when reading StorageMetadata from protobuff")
            size = ccl_leveldb.read_le_varint(stream)

            return cls(storage_key, timestamp, size, seq)


@dataclasses.dataclass(frozen=True)
class LocalStorageRecord:
    storage_key: str
    script_key: str
    value: str
    leveldb_seq_number: int
    is_live: bool

    @property
    def record_location(self) -> str:
        return f"Leveldb Seq: {self.leveldb_seq_number}"


class LocalStorageBatch:
    def __init__(self, meta: StorageMetadata, end_seq: int):
        self._meta = meta
        self._end = end_seq

    @property
    def storage_key(self) -> str:
        return self._meta.storage_key

    @property
    def timestamp(self) -> datetime.datetime:
        return self._meta.timestamp

    @property
    def start(self):
        return self._meta.leveldb_seq_number

    @property
    def end(self):
        return self._end

    def __repr__(self):
        return f"(storage_key={self.storage_key}, timestamp={self.timestamp}, start={self.start}, end={self.end})"


class LocalStoreDb:
    def __init__(self, in_dir: pathlib.Path):
        if not in_dir.is_dir():
            raise IOError("Input directory is not a directory")

        self._ldb = ccl_leveldb.RawLevelDb(in_dir)

        self._storage_details = {}  # storage_key: {seq_number: StorageMetadata}
        self._flat_items = []       # [StorageMetadata|LocalStorageRecord]   - used to batch items up
        self._records = {}          # storage_key: {script_key: {seq_number: LocalStorageRecord}}

        for record in self._ldb.iterate_records_raw():
            if record.user_key.startswith(_META_PREFIX) and record.state == ccl_leveldb.KeyState.Live:
                # Only live records for metadata - not sure what we can reliably infer from deleted keys
                storage_key = record.user_key.removeprefix(_META_PREFIX).decode(EIGHT_BIT_ENCODING)
                self._storage_details.setdefault(storage_key, {})
                metadata = StorageMetadata.from_protobuff(storage_key, record.value, record.seq)
                self._storage_details[storage_key][record.seq] = metadata
                self._flat_items.append(metadata)
            elif record.user_key.startswith(_RECORD_KEY_PREFIX):
                # We include deleted records here because we need them to build batches
                storage_key_raw, script_key_raw = record.user_key.removeprefix(_RECORD_KEY_PREFIX).split(b"\x00", 1)
                storage_key = storage_key_raw.decode(EIGHT_BIT_ENCODING)
                script_key = decode_string(script_key_raw)

                try:
                    value = decode_string(record.value) if record.state == ccl_leveldb.KeyState.Live else None
                except UnicodeDecodeError as e:
                    # Some sites play games to test the browser's capabilities like encoding half of a surrogate pair
                    print(f"Error decoding record value at seq no {record.seq}; "
                          f"{storage_key} {script_key}:  {record.value}")
                    continue

                self._records.setdefault(storage_key, {})
                self._records[storage_key].setdefault(script_key, {})

                ls_record = LocalStorageRecord(
                    storage_key, script_key, value, record.seq, record.state == ccl_leveldb.KeyState.Live)
                self._records[storage_key][script_key][record.seq] = ls_record
                self._flat_items.append(ls_record)

        self._storage_details = types.MappingProxyType(self._storage_details)
        self._records = types.MappingProxyType(self._records)

        self._all_storage_keys = frozenset(self._storage_details.keys() | self._records.keys())  # because deleted data.
        self._flat_items.sort(key=lambda x: x.leveldb_seq_number)

        # organise batches - this is made complex and slow by having to account for missing/deleted data
        # we're looking for a StorageMetadata followed by sequential (in terms of seq number) LocalStorageRecords
        # with the same storage key. Everything that falls within that chain can safely be considered a batch.
        # Any break in sequence numbers or storage key is a fail and can't be considered part of a batch.
        self._batches = {}
        current_meta: typing.Optional[StorageMetadata] = None
        current_end = 0
        for item in self._flat_items:  # pre-sorted
            if isinstance(item, LocalStorageRecord):
                if current_meta is None:
                    # no currently valid metadata so we can't attribute this record to anything
                    continue
                elif item.leveldb_seq_number - current_end != 1 or item.storage_key != current_meta.storage_key:
                    # this record breaks a chain, so bundle up what we have and clear everything out
                    self._batches[current_meta.leveldb_seq_number] = LocalStorageBatch(current_meta, current_end)
                    current_meta = None
                    current_end = 0
                else:
                    # contiguous and right storage key, include in the current chain
                    current_end = item.leveldb_seq_number
            elif isinstance(item, StorageMetadata):
                if current_meta is not None:
                    # this record breaks a chain, so bundle up what we have, set new start
                    self._batches[current_meta.leveldb_seq_number] = LocalStorageBatch(current_meta, current_end)
                current_meta = item
                current_end = item.leveldb_seq_number
            else:
                raise ValueError

        if current_meta is not None:
            self._batches[current_meta.leveldb_seq_number] = LocalStorageBatch(current_meta, current_end)

        self._batch_starts = tuple(sorted(self._batches.keys()))

    def iter_storage_keys(self) -> col_abc.Iterable[str]:
        yield from self._storage_details.keys()

    def contains_storage_key(self, storage_key: str) -> bool:
        return storage_key in self._all_storage_keys

    def iter_script_keys(self, storage_key: str) -> col_abc.Iterable[str]:
        if storage_key not in self._all_storage_keys:
            raise KeyError(storage_key)
        if storage_key not in self._records:
            raise StopIteration
        yield from self._records[storage_key].keys()

    def contains_script_key(self, storage_key: str, script_key: str) -> bool:
        return script_key in self._records.get(storage_key, {})

    def find_batch(self, seq: int) -> typing.Optional[LocalStorageBatch]:
        """
        Finds the batch that a record with the given sequence number belongs to
        :param seq: leveldb sequence id
        :return: the batch containing the given sequence number or None if no batch contains it
        """

        i = bisect.bisect_left(self._batch_starts, seq) - 1
        if i < 0:
            return None
        start = self._batch_starts[i]
        batch = self._batches[start]
        if batch.start <= seq <= batch.end:
            return batch
        else:
            return None

    def iter_all_records(self, include_deletions=False) -> col_abc.Iterable[LocalStorageRecord]:
        """
        :param include_deletions: if True, records related to deletions will be included
        (these will have None as values).
        :return: iterable of LocalStorageRecords
        """
        for storage_key, script_dict in self._records.items():
            for script_key, values in script_dict.items():
                for seq, value in values.items():
                    if value.is_live or include_deletions:
                        yield value

    def _iter_records_for_storage_key(
            self, storage_key: str, include_deletions=False) -> col_abc.Iterable[LocalStorageRecord]:
        """
        :param storage_key: storage key (host) for the records
        :param include_deletions: if True, records related to deletions will be included
        (these will have None as values).
        :return: iterable of LocalStorageRecords
        """
        if not self.contains_storage_key(storage_key):
            raise KeyError(storage_key)
        for script_key, values in self._records[storage_key].items():
            for seq, value in values.items():
                if value.is_live or include_deletions:
                    yield value

    def _search_storage_keys(self, storage_key: KeySearch) -> list[str]:
        if isinstance(storage_key, str):
            return [storage_key]
        elif isinstance(storage_key, re.Pattern):
            return [x for x in self._all_storage_keys if storage_key.search(x)]
        elif isinstance(storage_key, col_abc.Collection):
            return list(set(storage_key) & self._all_storage_keys)
        elif isinstance(storage_key, col_abc.Callable):
            return [x for x in self._all_storage_keys if storage_key(x)]
        else:
            raise TypeError(f"Unexpected type: {type(storage_key)} (expects: {KeySearch})")

    def iter_records_for_storage_key(
            self, storage_key: KeySearch, *,
            include_deletions=False, raise_on_no_result=True) -> col_abc.Iterable[LocalStorageRecord]:
        """
        :param storage_key: storage key (host) for the records. This can be one of: a single string;
        a collection of strings; a regex pattern; a function that takes a string (the host) and returns a bool.
        :param include_deletions: if True, records related to deletions will be included
        :param raise_on_no_result: if True (the default) if no matching storage keys are found, raise a KeyError
        (these will have None as values).
        :return: iterable of LocalStorageRecords
        """
        if isinstance(storage_key, str):
            if raise_on_no_result and not self.contains_storage_key(storage_key):
                raise KeyError(storage_key)
            yield from self._iter_records_for_storage_key(storage_key, include_deletions)
        elif isinstance(storage_key, re.Pattern):
            matched_keys = self._search_storage_keys(storage_key)
            if raise_on_no_result and not matched_keys:
                raise KeyError(f"Pattern: {storage_key.pattern}")
            for key in matched_keys:
                yield from self._iter_records_for_storage_key(key, include_deletions)
        elif isinstance(storage_key, col_abc.Collection):
            matched_keys = self._search_storage_keys(storage_key)
            if raise_on_no_result and not matched_keys:
                raise KeyError(storage_key)
            for key in matched_keys:
                yield from self._iter_records_for_storage_key(key, include_deletions)
        elif isinstance(storage_key, col_abc.Callable):
            matched_keys = self._search_storage_keys(storage_key)
            if raise_on_no_result and not matched_keys:
                raise KeyError(storage_key)
            for key in matched_keys:
                yield from self._iter_records_for_storage_key(key, include_deletions)
        else:
            raise TypeError(f"Unexpected type for storage key: {type(storage_key)} (expects: {KeySearch})")

    def _iter_records_for_script_key(
            self, storage_key: str, script_key: str, include_deletions=False) -> col_abc.Iterable[LocalStorageRecord]:
        """
        :param storage_key: storage key (host) for the records
        :param script_key: script defined key for the records
        :param include_deletions: if True, records related to deletions will be included
        :return: iterable of LocalStorageRecords
        """
        if not self.contains_script_key(storage_key, script_key):
            raise KeyError((storage_key, script_key))
        for seq, value in self._records[storage_key][script_key].items():
            if value.is_live or include_deletions:
                yield value

    def iter_records_for_script_key(
        self, storage_key: KeySearch, script_key: KeySearch, *,
            include_deletions=False, raise_on_no_result=True) -> col_abc.Iterable[LocalStorageRecord]:
        """
        :param storage_key: storage key (host) for the records. This can be one of: a single string;
        a collection of strings; a regex pattern; a function that takes a string and returns a bool.
        :param script_key: script defined key for the records. This can be one of: a single string;
        a collection of strings; a regex pattern; a function that takes a string and returns a bool.
        :param include_deletions: if True, records related to deletions will be included
        :param raise_on_no_result: if True (the default) if no matching storage keys are found, raise a KeyError
        (these will have None as values).
        :return: iterable of LocalStorageRecords
        """

        if isinstance(storage_key, str) and isinstance(script_key, str):
            if raise_on_no_result and not self.contains_script_key(storage_key, script_key):
                raise KeyError((storage_key, script_key))
            yield from self._iter_records_for_script_key(storage_key, script_key, include_deletions=include_deletions)
        else:
            matched_storage_keys = self._search_storage_keys(storage_key)
            if raise_on_no_result and not matched_storage_keys:
                raise KeyError((storage_key, script_key))

            yielded = False
            for matched_storage_key in matched_storage_keys:
                if isinstance(script_key, str):
                    matched_script_keys = [script_key]
                elif isinstance(script_key, re.Pattern):
                    matched_script_keys = [x for x in self._records[matched_storage_key].keys() if script_key.search(x)]
                elif isinstance(script_key, col_abc.Collection):
                    script_key_set = set(script_key)
                    matched_script_keys = list(self._records[matched_storage_key].keys() & script_key_set)
                elif isinstance(script_key, col_abc.Callable):
                    matched_script_keys = [x for x in self._records[matched_storage_key].keys() if script_key(x)]
                else:
                    raise TypeError(f"Unexpected type for script key: {type(script_key)} (expects: {KeySearch})")

                for key in matched_script_keys:
                    for seq, value in self._records[matched_storage_key][key].items():
                        if value.is_live or include_deletions:
                            yielded = True
                            yield value

            if not yielded:
                raise KeyError((storage_key, script_key))

    def iter_metadata(self) -> col_abc.Iterable[StorageMetadata]:
        """
        :return: iterable of StorageMetaData
        """
        for meta in self._flat_items:
            if isinstance(meta, StorageMetadata):
                yield meta

    def iter_metadata_for_storage_key(self, storage_key: str) -> col_abc.Iterable[StorageMetadata]:
        """
        :param storage_key: storage key (host) for the metadata
        :return: iterable of StorageMetadata
        """
        if storage_key not in self._all_storage_keys:
            raise KeyError(storage_key)
        if storage_key not in self._storage_details:
            return None
        for seq, meta in self._storage_details[storage_key].items():
            yield meta

    def iter_batches(self) -> col_abc.Iterable[LocalStorageBatch]:
        yield from self._batches.values()

    def close(self):
        self._ldb.close()

    def __contains__(self, item: typing.Union[str, tuple[str, str]]) -> bool:
        """
        :param item: either the host as a str or a tuple of the host and a key (both str)
        :return: if item is a str, returns true if that host is present, if item is a tuple of (str, str), returns True
            if that host and key pair are present
        """

        if isinstance(item, str):
            return item in self._all_storage_keys
        elif isinstance(item, tuple) and len(item) == 2:
            host, key = item
            return host in self._all_storage_keys and key in self._records[host]
        else:
            raise TypeError("item must be a string or a tuple of (str, str)")

    def __iter__(self):
        """
        iterates the hosts (storage keys) present
        """
        yield from self._all_storage_keys

    def __enter__(self) -> "LocalStoreDb":
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


def main(args):
    in_ldb_path = pathlib.Path(args[0])
    local_store = LocalStoreDb(in_ldb_path)

    for rec in local_store.iter_all_records():
        batch = local_store.find_batch(rec.leveldb_seq_number)
        print(rec, batch)


if __name__ == '__main__':
    main(sys.argv[1:])




"""
Copyright 2022, CCL Forensics

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

__version__ = "0.3"
__description__ = "A module for reading downloads from the Chrome/Chromium shared_proto_db leveldb data store"
__contact__ = "Alex Caithness"

import datetime
import io
import os
import pathlib
import sys
import typing

from .storage_formats import ccl_leveldb
from .serialization_formats import ccl_protobuff as pb
from .download_common import Download

CHROME_EPOCH = datetime.datetime(1601, 1, 1, 0, 0, 0)


def chrome_milli_time(milliseconds: typing.Optional[int], allow_none=True) -> typing.Optional[datetime.datetime]:
    if milliseconds is not None:
        if milliseconds == 0xffffffffffffffff:
            return CHROME_EPOCH
        else:
            return CHROME_EPOCH + datetime.timedelta(milliseconds=milliseconds)
    elif allow_none:
        return None
    raise ValueError("milliseconds cannot be None")


def read_datetime(stream) -> typing.Optional[datetime.datetime]:
    return chrome_milli_time(pb.read_le_varint(stream))


# https://source.chromium.org/chromium/chromium/src/+/main:components/download/database/proto/download_entry.proto;l=86

HttpRequestHeader_Structure = {
    1: pb.ProtoDecoder("key", pb.read_string),
    2: pb.ProtoDecoder("value", pb.read_string)
}

ReceivedSlice_Structure = {
    1: pb.ProtoDecoder("offset", pb.read_le_varint),
    2: pb.ProtoDecoder("received_bytes", pb.read_le_varint),
    3: pb.ProtoDecoder("finished", lambda x: pb.read_le_varint(x) != 0)
}

InProgressInfo_Structure = {
    1: pb.ProtoDecoder("url_chain", pb.read_string),  # string
    2: pb.ProtoDecoder("referrer_url", pb.read_string),  # string
    3: pb.ProtoDecoder("site_url", pb.read_string),  # string  // deprecated
    4: pb.ProtoDecoder("tab_url", pb.read_string),  # string
    5: pb.ProtoDecoder("tab_referrer_url", pb.read_string),  # string
    6: pb.ProtoDecoder("fetch_error_body", lambda x: pb.read_le_varint(x) != 0),  # bool
    7: pb.ProtoDecoder("request_headers", lambda x: pb.read_embedded_protobuf(x, HttpRequestHeader_Structure, True)),  # HttpRequestHeader
    8: pb.ProtoDecoder("etag", pb.read_string),  # string
    9: pb.ProtoDecoder("last_modified", pb.read_string),  # string
    10: pb.ProtoDecoder("total_bytes", pb.read_le_varint),  # int64:
    11: pb.ProtoDecoder("mime_type", pb.read_string),  # string
    12: pb.ProtoDecoder("original_mime_type", pb.read_string),  # string
    13: pb.ProtoDecoder("current_path", pb.read_blob),  # bytes  // Serialized pickles to support string16: TODO
    14: pb.ProtoDecoder("target_path", pb.read_blob),  # bytes   // Serialized pickles to support string16: TODO
    15: pb.ProtoDecoder("received_bytes", pb.read_le_varint),  # int64:
    16: pb.ProtoDecoder("start_time", read_datetime),  # int64:
    17: pb.ProtoDecoder("end_time", read_datetime),  # int64:
    18: pb.ProtoDecoder("received_slices", lambda x: pb.read_embedded_protobuf(x, ReceivedSlice_Structure, True)),  # ReceivedSlice
    19: pb.ProtoDecoder("hash", pb.read_blob),  # string
    20: pb.ProtoDecoder("transient", lambda x: pb.read_le_varint(x) != 0),  # bool
    21: pb.ProtoDecoder("state", pb.read_le_varint32),  # int32:
    22: pb.ProtoDecoder("danger_type", pb.read_le_varint32),  # int32:
    23: pb.ProtoDecoder("interrupt_reason", pb.read_le_varint32),  # int32:
    24: pb.ProtoDecoder("paused", lambda x: pb.read_le_varint(x) != 0),  # bool
    25: pb.ProtoDecoder("metered", lambda x: pb.read_le_varint(x) != 0),  # bool
    26: pb.ProtoDecoder("bytes_wasted", pb.read_le_varint),  # int64:
    27: pb.ProtoDecoder("auto_resume_count", pb.read_le_varint32),  # int32:
    # 28: pb.ProtoDecoder("download_schedule", None)  # DownloadSchedule  // // Deprecated.
    # 29: pb.ProtoDecoder("reroute_info", pb),  # enterprise_connectors.DownloadItemRerouteInfo TODO
    30: pb.ProtoDecoder("credentials_mode", pb.read_le_varint32),  # int32:  // network::mojom::CredentialsMode
    31: pb.ProtoDecoder("range_request_from", pb.read_le_varint),  # int64:
    32: pb.ProtoDecoder("range_request_to", pb.read_le_varint),  # int64:
    33: pb.ProtoDecoder("serialized_embedder_download_data", pb.read_string)  # string
}

DownloadInfo_structure = {
    1: pb.ProtoDecoder("guid", pb.read_string),
    2: pb.ProtoDecoder("id", pb.read_le_varint32),
    # 3 UkmInfo
    4: pb.ProtoDecoder("in_progress_info", lambda x: pb.read_embedded_protobuf(x, InProgressInfo_Structure, True))
}

DownloadDbEntry_structure = {
    1: pb.ProtoDecoder("download_info", lambda x: pb.read_embedded_protobuf(x, DownloadInfo_structure, True))
}


def read_downloads(
        shared_proto_db_folder: typing.Union[str, os.PathLike],
        *, handle_errors=False, utf16_paths=True) -> typing.Iterator[Download]:
    ldb_path = pathlib.Path(shared_proto_db_folder)
    with ccl_leveldb.RawLevelDb(ldb_path) as ldb:
        for rec in ldb.iterate_records_raw():
            if rec.state != ccl_leveldb.KeyState.Live:
                continue

            key = rec.user_key
            record_type, specific_key = key.split(b"_", 1)
            if record_type == b"21":
                with io.BytesIO(rec.value) as f:
                    obj = pb.ProtoObject(
                        0xa, "root", pb.read_protobuff(f, DownloadDbEntry_structure, use_friendly_tag=True))
                try:
                    download = Download.from_pb(rec.seq, obj, target_path_is_utf_16=utf16_paths)
                except ValueError as ex:
                    print(f"Error reading a download: {ex}", file=sys.stderr)
                    if handle_errors:
                        continue
                    else:
                        raise

                yield download


def report_downloads(
        shared_proto_db_folder: typing.Union[str, os.PathLike],
        out_csv_path: typing.Union[str, os.PathLike], utf16_paths=True):

    with pathlib.Path(out_csv_path).open("tx", encoding="utf-8", newline="") as out:
        writer = csv.writer(out, csv.excel, quoting=csv.QUOTE_ALL, quotechar="\"", escapechar="\\")
        writer.writerow([
            "seq no",
            "guid",
            "start time",
            "end time",
            "tab url",
            "tab referrer url",
            "download url chain",
            "target path",
            "hash",
            "total bytes",
            "mime type",
            "original mime type"
        ])
        for download in read_downloads(shared_proto_db_folder, handle_errors=True, utf16_paths=utf16_paths):
            writer.writerow([
                str(download.level_db_seq_no),
                str(download.guid),
                download.start_time,
                download.end_time,
                download.tab_url,
                download.tab_referrer_url,
                " -> ".join(download.url_chain),
                download.target_path,
                download.hash,
                download.total_bytes,
                download.mime_type,
                download.original_mime_type
            ])


if __name__ == '__main__':
    import csv
    if len(sys.argv) < 3:
        print(f"USAGE: {pathlib.Path(sys.argv[0]).name} <shared_proto_db folder> <out.csv> [-u8]")
        print()
        print("-u8\tutf-8 target paths (use this if target paths appear garbled in the output)")
        print()
        exit(1)
    report_downloads(sys.argv[1], sys.argv[2], "-u8" not in sys.argv[3:])



"""
Tool routing policy for unified local-fast + official Linear MCP access.
"""

from __future__ import annotations

import logging
import os
import threading
import time
from typing import Any

from . import local_handlers
from .official_session import OfficialMcpSessionManager, OfficialToolError
from .reader import LinearLocalReader

logger = logging.getLogger(__name__)

WRITE_TOOL_PREFIXES = (
    "create_",
    "update_",
    "delete_",
    "archive_",
    "unarchive_",
    "set_",
    "add_",
    "remove_",
    "move_",
)


class ToolRouter:
    """Routes tool calls between local cache handlers and official MCP."""

    def __init__(
        self,
        reader: LinearLocalReader,
        official: OfficialMcpSessionManager,
        coherence_window_seconds: int | None = None,
    ):
        self._reader = reader
        self._official = official
        self._coherence_window_seconds = coherence_window_seconds or int(
            os.getenv("LINEAR_FAST_COHERENCE_WINDOW_SECONDS", "30")
        )
        self._remote_reads_until = 0.0
        self._state_lock = threading.RLock()

    def _mark_recent_write(self) -> None:
        with self._state_lock:
            self._remote_reads_until = time.time() + self._coherence_window_seconds

    def _read_remote_first(self) -> bool:
        with self._state_lock:
            return time.time() < self._remote_reads_until

    def _is_probable_write_tool(self, tool_name: str) -> bool:
        if tool_name in local_handlers.LOCAL_READ_HANDLERS:
            return False
        return tool_name.startswith(WRITE_TOOL_PREFIXES)

    def _call_local(
        self,
        tool_name: str,
        arguments: dict[str, Any],
        *,
        allow_degraded: bool = False,
    ) -> Any:
        handler = local_handlers.LOCAL_READ_HANDLERS.get(tool_name)
        if handler is None:
            raise local_handlers.LocalFallbackRequested(
                "unsupported_tool", f"tool '{tool_name}' not implemented in local cache"
            )

        if self._reader.is_degraded() and not allow_degraded:
            raise local_handlers.LocalFallbackRequested(
                "degraded_local", "local cache is degraded"
            )

        return handler(self._reader, **arguments)

    def call_official(self, tool_name: str, arguments: dict[str, Any] | None = None) -> Any:
        result = self._official.call_tool(tool_name, arguments or {})
        if self._is_probable_write_tool(tool_name):
            self._mark_recent_write()
        return result

    def call_read(self, tool_name: str, arguments: dict[str, Any] | None = None) -> Any:
        args = arguments or {}
        remote_error: OfficialToolError | None = None

        if self._read_remote_first():
            try:
                return self.call_official(tool_name, args)
            except OfficialToolError as exc:
                if exc.code == "official_tool_error":
                    raise
                logger.warning("Remote-first read failed for %s, falling back to local", tool_name)
                remote_error = exc

        try:
            return self._call_local(tool_name, args)
        except local_handlers.LocalFallbackRequested as local_exc:
            if remote_error is not None:
                if local_exc.code == "degraded_local":
                    logger.warning(
                        "Returning stale local for %s because remote failed during remote-first window",
                        tool_name,
                    )
                    return self._call_local(tool_name, args, allow_degraded=True)
                raise remote_error

            try:
                return self.call_official(tool_name, args)
            except OfficialToolError as official_exc:
                if official_exc.code == "official_tool_error":
                    raise
                if local_exc.code == "degraded_local":
                    # When remote is unavailable, stale local read is better than hard failure.
                    logger.warning(
                        "Returning stale local for %s because remote is unavailable and local is degraded",
                        tool_name,
                    )
                    return self._call_local(tool_name, args, allow_degraded=True)
                raise
        except Exception:
            logger.exception("Unexpected local error for %s", tool_name)
            return self.call_official(tool_name, args)

    def refresh_local_cache(self) -> dict[str, Any]:
        self._reader.refresh_cache(force=True)
        return self._reader.get_health()

    def get_health(self) -> dict[str, Any]:
        return {
            "local": self._reader.get_health(),
            "official": self._official.get_health(),
            "remoteReadUntil": self._remote_reads_until,
            "coherenceWindowSeconds": self._coherence_window_seconds,
        }



"""
Copyright 2022, CCL Forensics

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import datetime
import enum
import io
import os
import struct
import sys
import pathlib
import dataclasses
import typing

from .storage_formats import ccl_leveldb
from .serialization_formats import ccl_blink_value_deserializer, ccl_v8_value_deserializer, ccl_protobuff as pb

__version__ = "0.3"
__description__ = "Library for reading Chrome/Chromium notifications (Platform Notifications)"
__contact__ = "Alex Caithness"

# See content/browser/notifications/notification_database.cc
#   and content/browser/notifications/notification_database_data.proto

EPOCH = datetime.datetime(1601, 1, 1)


# stolen from ccl_chromium_indexeddb 20230907
@dataclasses.dataclass(frozen=True)
class BlinkTrailer:
    # third_party/blink/renderer/bindings/core/v8/serialization/trailer_reader.h
    offset: int
    length: int

    TRAILER_SIZE: typing.ClassVar[int] = 13
    MIN_WIRE_FORMAT_VERSION_FOR_TRAILER: typing.ClassVar[int] = 21

    @classmethod
    def from_buffer(cls, buffer, trailer_offset: int):
        tag, offset, length = struct.unpack(">cQI", buffer[trailer_offset: trailer_offset + BlinkTrailer.TRAILER_SIZE])
        if tag != ccl_blink_value_deserializer.Constants.tag_kTrailerOffsetTag:
            raise ValueError(
                f"Trailer doesn't start with kTrailerOffsetTag "
                f"(expected: 0x{ccl_blink_value_deserializer.Constants.tag_kTrailerOffsetTag.hex()}; "
                f"got: 0x{tag.hex()}")

        return BlinkTrailer(offset, length)


class ClosedReason(enum.IntEnum):
    USER = 0
    DEVELOPER = 1
    UNKNOWN = 2


class ActionType(enum.IntEnum):
    BUTTON = 0
    TEXT = 1


class Direction(enum.IntEnum):
    LEFT_TO_RIGHT = 0
    RIGHT_TO_LEFT = 1
    AUTO = 2


def read_datetime(stream):
    ms = pb.read_le_varint(stream)
    return EPOCH + datetime.timedelta(microseconds=ms)


NotificationAction_Structure = {
    1: pb.ProtoDecoder("action", pb.read_string),
    2: pb.ProtoDecoder("title", pb.read_string),
    3: pb.ProtoDecoder("icon", pb.read_string),
    4: pb.ProtoDecoder("type", lambda x: ActionType(pb.read_le_varint(x))),
    5: pb.ProtoDecoder("placeholder", pb.read_string),
}

NotificationData_Structure = {
    1: pb.ProtoDecoder("title", pb.read_string),
    2: pb.ProtoDecoder("closed_reason", lambda x: Direction(pb.read_le_varint(x))),
    3: pb.ProtoDecoder("lang", pb.read_string),
    4: pb.ProtoDecoder("body", pb.read_string),
    5: pb.ProtoDecoder("tag", pb.read_string),
    6: pb.ProtoDecoder("icon", pb.read_string),
    7: pb.ProtoDecoder("silent", lambda x: pb.read_le_varint(x) != 0),
    8: pb.ProtoDecoder("data", pb.read_blob),
    9: pb.ProtoDecoder("vibration", pb.read_blob),
    10: pb.ProtoDecoder(
        "actions", lambda x: pb.read_embedded_protobuf(x, NotificationAction_Structure, use_friendly_tag=True)),
    11: pb.ProtoDecoder("require_interaction", lambda x: pb.read_le_varint(x) != 0),
    12: pb.ProtoDecoder("timestamp", read_datetime),
    13: pb.ProtoDecoder("renotify", lambda x: pb.read_le_varint(x) != 0),
    14: pb.ProtoDecoder("badge", pb.read_string),
    15: pb.ProtoDecoder("image", pb.read_string),
    16: pb.ProtoDecoder("show_trigger_timestamp", read_datetime),
}

NotificationDatabaseDataProto_Structure = {
    1: pb.ProtoDecoder("persistent_notification_id", pb.read_le_varint),
    2: pb.ProtoDecoder("origin", pb.read_string),
    3: pb.ProtoDecoder("service_worker_registration_id", pb.read_le_varint),
    4: pb.ProtoDecoder(
        "notification_data", lambda x: pb.read_embedded_protobuf(x, NotificationData_Structure, use_friendly_tag=True)),
    5: pb.ProtoDecoder("notification_id", pb.read_string),
    6: pb.ProtoDecoder("replaced_existing_notification", lambda x: pb.read_le_varint(x) != 0),
    7: pb.ProtoDecoder("num_clicks", pb.read_le_varint32),
    8: pb.ProtoDecoder("num_action_button_clicks", pb.read_le_varint32),
    9: pb.ProtoDecoder("creation_time_millis", read_datetime),
    10: pb.ProtoDecoder("time_until_first_click_millis", pb.read_le_varint),
    11: pb.ProtoDecoder("time_until_last_click_millis", pb.read_le_varint),
    12: pb.ProtoDecoder("time_until_close_millis", pb.read_le_varint),
    13: pb.ProtoDecoder("closed_reason", lambda x: ClosedReason(pb.read_le_varint(x))),
    14: pb.ProtoDecoder("has_triggered", lambda x: pb.read_le_varint(x) != 0),
    15: pb.ProtoDecoder("is_shown_by_browser", lambda x: pb.read_le_varint(x) != 0),
}


@dataclasses.dataclass(frozen=True)
class LevelDbInfo:
    user_key: bytes
    origin_file: os.PathLike
    seq_no: int


@dataclasses.dataclass(frozen=True)
class NotificationAction:
    action: typing.Optional[str]
    title: typing.Optional[str]
    icon: typing.Optional[str]
    action_type: typing.Optional[ActionType]
    placeholder: typing.Optional[str]


@dataclasses.dataclass(frozen=True)
class ChromiumNotification:
    level_db_info: LevelDbInfo
    origin: str
    persistent_notification_id: int
    notification_id: str
    title: typing.Optional[str]
    body: typing.Optional[str]
    data: typing.Optional[typing.Any]
    timestamp: datetime.datetime
    creation_time: datetime.datetime  # from creation_time_millis
    closed_reason: ClosedReason
    time_until_first_click_millis: int
    time_until_last_click_millis: int
    time_until_close_millis: int

    tag: typing.Optional[str]
    image: typing.Optional[str]
    icon: typing.Optional[str]
    badge: typing.Optional[str]

    actions: typing.Optional[typing.Iterable[NotificationAction]]


class NotificationReader:
    def __init__(self, notification_input_path: pathlib.Path):
        self._db = ccl_leveldb.RawLevelDb(notification_input_path)

    def close(self):
        self._db.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._db.close()

    def read_notifications(self) -> typing.Iterable[ChromiumNotification]:
        blink_deserializer = ccl_blink_value_deserializer.BlinkV8Deserializer()
        for record in self._db.iterate_records_raw():
            if record.state != ccl_leveldb.KeyState.Live:
                continue

            key = record.user_key.decode("utf-8")
            record_type, key_info = key.split(":", 1)
            origin, key_id = key_info.split("\0", 1)
            level_db_info = LevelDbInfo(record.user_key, record.origin_file, record.seq)
            if record_type == "DATA":
                with io.BytesIO(record.value) as stream:
                    root = pb.ProtoObject(
                        0x2,
                        "root",
                        pb.read_protobuff(stream, NotificationDatabaseDataProto_Structure, use_friendly_tag=True))

                data = root.only("notification_data").only("data").value
                if data:
                    if data[0] != 0xff:
                        print(key)
                        print(data)
                        raise ValueError("Missing blink tag at the start of data")
                    blink_version, blink_version_bytes = pb._read_le_varint(io.BytesIO(data[1:]))
                    data_start = 1 + len(blink_version_bytes)
                    if blink_version >= BlinkTrailer.MIN_WIRE_FORMAT_VERSION_FOR_TRAILER:
                        trailer = BlinkTrailer.from_buffer(data, data_start)  # TODO: do something with the trailer?
                        data_start += BlinkTrailer.TRAILER_SIZE

                    with io.BytesIO(data[data_start:]) as obj_raw:
                        try:
                            deserializer = ccl_v8_value_deserializer.Deserializer(
                                obj_raw, host_object_delegate=blink_deserializer.read)
                        except ValueError:
                            print("Error record:")
                            print(level_db_info, key)
                            raise
                        data = deserializer.read()

                yield ChromiumNotification(
                    level_db_info,
                    root.only("origin").value,
                    root.only("persistent_notification_id").value,
                    root.only("notification_id").value,
                    root.only("notification_data").only("title").value,
                    root.only("notification_data").only("body").value,
                    data,
                    root.only("notification_data").only("timestamp").value,
                    root.only("creation_time_millis").value,
                    root.only("closed_reason").value,
                    root.only("time_until_first_click_millis").value,
                    root.only("time_until_last_click_millis").value,
                    root.only("time_until_close_millis").value,
                    root.only("notification_data").only("tag").value,
                    root.only("notification_data").only("image").value,
                    root.only("notification_data").only("icon").value,
                    root.only("notification_data").only("badge").value,
                    tuple(
                        NotificationAction(
                            x.only("action").value,
                            x.only("title").value,
                            x.only("icon").value,
                            x.only("type").value,
                            x.only("placeholder").value
                        )
                        for x in root["notification_data"][0]["actions"])
                )


if __name__ == '__main__':
    if len(sys.argv) < 2:
        print(f"USAGE: {pathlib.Path(sys.argv[0]).name} <Platform Notifications Folder>")
        exit(1)

    _reader = NotificationReader(pathlib.Path(sys.argv[1]))
    _blink_deserializer = ccl_blink_value_deserializer.BlinkV8Deserializer()
    for notification in _reader.read_notifications():
        print(notification)



"""
Copyright 2022, CCL Forensics

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

__version__ = "0.8"
__description__ = "Library for reading Chrome/Chromium File System API data"
__contact__ = "Alex Caithness"

import dataclasses
import os
import sys
import pathlib
import datetime
import re
import typing
import types
import functools

from .storage_formats import ccl_leveldb
from .serialization_formats.ccl_easy_chromium_pickle import EasyPickleIterator


@dataclasses.dataclass(frozen=True)
class FileInfo:
    _owner: "FileSystem" = dataclasses.field(repr=False)
    origin: str
    folder_id: str
    is_persistent: bool
    seq_no: int
    file_id: int
    parent_id: int
    data_path: str
    name: str
    timestamp: datetime.datetime

    @classmethod
    def from_pickle(
            cls, owner: "FileSystem", origin: str, folder_id: str, is_persistent: bool,
            seq_no: int, file_id: int, data: bytes):
        with EasyPickleIterator(data) as reader:
            parent_id = reader.read_uint64()
            data_path = reader.read_string()
            name = reader.read_string()
            timestamp = reader.read_datetime()

        return cls(owner, origin, folder_id, is_persistent, seq_no, file_id, parent_id, data_path, name, timestamp)

    def get_local_storage_path(self) -> pathlib.Path:
        return self._owner.get_local_path_for_fileinfo(self)

    @property
    def is_stored_locally(self) -> bool:
        return self.get_local_storage_path().exists()


class OriginStorage:
    def __init__(
            self,
            owner: "FileSystem",
            origin: str,
            folder_id: str,
            persistent_files: typing.Optional[typing.Mapping[int, FileInfo]],
            persistent_deleted_file_ids: typing.Optional[typing.Iterable[int]],
            temporary_files: typing.Optional[typing.Mapping[int, FileInfo]],
            temporary_deleted_file_ids: typing.Optional[typing.Iterable[int]]):
        self._owner = owner
        self._origin = origin
        self._folder_id = folder_id
        self._persistent_files = types.MappingProxyType(persistent_files or {})
        self._persistent_deleted_file_ids = set(persistent_deleted_file_ids or [])
        self._temporary_files = types.MappingProxyType(temporary_files or {})
        self._temporary_deleted_file_ids = set(temporary_deleted_file_ids or [])

        self._persistent_file_listing_lookup = types.MappingProxyType(self._make_file_listing_lookup(True))
        self._temporary_file_listing_lookup = types.MappingProxyType(self._make_file_listing_lookup(False))

        self._file_listing_lookup_reverse: dict[str, list[str]] = {}
        for k, v in self._persistent_file_listing_lookup.items():
            self._file_listing_lookup_reverse.setdefault(v, [])
            self._file_listing_lookup_reverse[v].append(f"p_{k}")

        for k, v in self._temporary_file_listing_lookup.items():
            self._file_listing_lookup_reverse.setdefault(v, [])
            self._file_listing_lookup_reverse[v].append(f"t_{k}")
        self._file_listing_lookup_reverse = types.MappingProxyType(
            self._file_listing_lookup_reverse)

    def _make_file_listing_lookup(self, persistent=True) -> dict[int, str]:
        files = self._persistent_files if persistent else self._temporary_files
        file_listing_lookup: dict[int, str] = {}

        for file_info in files.values():
            if not file_info.data_path:
                continue
            path_parts = []
            current = file_info
            while current.file_id != current.parent_id:
                path_parts.insert(0, current.name)
                current = files.get(current.parent_id, "<MISSING PATH SEGMENT>")

            path_parts.insert(0, "p" if persistent else "t")
            # path_parts.insert(0, self._origin)
            # path_parts.insert(0, "")
            file_listing_lookup[file_info.file_id] = "/".join(path_parts)

        return file_listing_lookup

    def get_file_listing(self) -> typing.Iterable[tuple[str, FileInfo]]:
        for file_id in self._persistent_file_listing_lookup:
            yield self._persistent_file_listing_lookup[file_id], self._persistent_files[file_id]
        for file_id in self._temporary_file_listing_lookup:
            yield self._temporary_file_listing_lookup[file_id], self._temporary_files[file_id]

    def _get_file_info_from_path(self, path) -> typing.Iterable[FileInfo]:
        file_keys = self._file_listing_lookup_reverse[str(path)]
        for key in file_keys:
            p_or_t, file_id = key.split("_", 1)
            yield self._persistent_files[int(file_id)] if p_or_t == "p" else self._temporary_files[int(file_id)]


class FileSystem:
    def __init__(self, path: typing.Union[os.PathLike, str]):
        """
        Constructor for the File System API access (the entry point for most processing scripts)
        :param path: the path of the File System API storage
        """
        self._root = pathlib.Path(path)
        self._origins = self._get_origins()
        self._origins_reverse = {}
        for origin, folders in self._origins.items():
            for folder in folders:
                self._origins_reverse[folder] = origin

    def _get_origins(self) -> dict[str, tuple]:
        result = {}
        with ccl_leveldb.RawLevelDb(self._root / "Origins") as db:
            for record in db.iterate_records_raw():
                if record.state != ccl_leveldb.KeyState.Live:
                    continue
                if record.user_key.startswith(b"ORIGIN:"):
                    _, origin = record.user_key.split(b":", 1)
                    origin = origin.decode("utf-8")
                    result.setdefault(origin, [])
                    result[origin].append(record.value.decode("utf-8"))

        return {k: tuple(v) for (k, v) in result.items()}

    def get_origins(self) -> typing.Iterable[str]:
        """
        Yields the origins for this File System API
        :return: Yields the origins in this File System API
        """
        yield from self._origins.keys()

    def get_folders_for_origin(self, origin) -> tuple[str, ...]:
        """
        Returns the folder ids which are used by the origin (host/domain)
        :param origin:
        :return: a tuple of strings which are the folder id(s) for this origin
        """
        return self._origins[origin]

    def get_storage_for_folder(self, folder_id) -> OriginStorage:
        """
        Get the OriginStorage object for the folder
        :param folder_id: a folder id (such as those returned by get_folders_for_origin)
        :return: OriginStorage for the folder_id
        """
        return self._build_file_graph(folder_id)

    @functools.cache
    def _build_file_graph(self, folder_id) -> OriginStorage:
        persistent_files: typing.Optional[dict[int, FileInfo]] = {}
        persistent_deleted_files: typing.Optional[dict[int, int]] = {}  # file_id: seq_no
        temporary_files: typing.Optional[dict[int, FileInfo]] = {}
        temporary_deleted_files: typing.Optional[dict[int, int]] = {}  # file_id: seq_no

        origin = self._origins_reverse[folder_id]

        for p_or_t in ("p", "t"):
            db_path = self._root / folder_id / p_or_t / "Paths"
            if not db_path.exists():
                continue
            files: dict[int, FileInfo] = persistent_files if p_or_t == "p" else temporary_files
            deleted_files: dict[int, int] = persistent_deleted_files if p_or_t == "p" else temporary_deleted_files
            with ccl_leveldb.RawLevelDb(db_path) as db:
                # TODO: we can infer file modified (created?) times using the parent's modified times maybe
                for record in db.iterate_records_raw():
                    if re.match(b"[0-9]+", record.user_key) is not None:
                        if record.state == ccl_leveldb.KeyState.Live:
                            file_id = int(record.user_key.decode("utf-8"))
                            file_info = FileInfo.from_pickle(
                                self, origin, folder_id, p_or_t == "p", record.seq, file_id, record.value)

                            # undelete a file if more recent than deletion record:
                            if file_id in deleted_files and deleted_files[file_id] < file_info.seq_no:
                                deleted_files.pop(file_id)

                            if old_file_info := files.get(file_id):
                                if old_file_info.seq_no < file_info.seq_no:
                                    # TODO: any reason to keep older records (other than for the timestamps as above?)
                                    files[file_id] = file_info
                            else:
                                files[file_id] = file_info
                        else:
                            if old_file_info := files.get(file_id):
                                if old_file_info.seq_no < record.seq:
                                    deleted_files[file_id] = record.seq
                            else:
                                deleted_files[file_id] = record.seq

        return OriginStorage(
            self, origin, folder_id,
            persistent_files, persistent_deleted_files.keys(),
            temporary_files, temporary_deleted_files.keys())

    def get_local_path_for_fileinfo(self, file_info: FileInfo):
        """
        Returns the path on the local file system for the FilInfo object
        :param file_info:
        :return: the path on the local file system for the FilInfo object
        """
        path = self._root / file_info.folder_id / ("p" if file_info.is_persistent else "t") / file_info.data_path
        return path

    def get_file_stream_for_fileinfo(self, file_info: FileInfo) -> typing.Optional[typing.BinaryIO]:
        """
        Returns a file object from the local file system for the FilInfo object
        :param file_info:
        :return: a file object from the local file system for the FilInfo object
        """
        path = self.get_local_path_for_fileinfo(file_info)
        if path.exists():
            return path.open("rb")
        return None


class FileSystemUtils:
    @staticmethod
    def print_origin_to_folder(fs_folder: typing.Union[os.PathLike, str]) -> None:
        """
        utility function to print out origins in the File System API to their folders
        :param fs_folder: the path of the File System API storage
        :return: None
        """
        fs = FileSystem(fs_folder)
        for origin in sorted(fs.get_origins()):
            print(f"{origin}: {','.join(fs.get_folders_for_origin(origin))}")

    @staticmethod
    def print_folder_to_origin(fs_folder: typing.Union[os.PathLike, str]) -> None:
        """
        utility function to print out folders in the File System API to their Origin
        :param fs_folder: the path of the File System API storage
        :return: None
        """
        fs = FileSystem(fs_folder)
        result = {}
        for origin in fs.get_origins():
            for folder in fs.get_folders_for_origin(origin):
                result[folder] = origin

        for folder in sorted(result.keys()):
            print(f"{folder}: {result[folder]}")

    @staticmethod
    def print_all_files(fs_folder: typing.Union[os.PathLike, str]) -> None:
        """
        utility function to print out all files in the File System API
        :param fs_folder: the path of the File System API storage
        :return: None
        """
        fs = FileSystem(fs_folder)
        for origin in sorted(fs.get_origins()):
            for folder in fs.get_folders_for_origin(origin):
                storage = fs.get_storage_for_folder(folder)
                for file_path, file_info in storage.get_file_listing():
                    print("/".join([origin, file_path]))


if __name__ == "__main__":
    FileSystemUtils.print_all_files(sys.argv[1])


